#
# BFD transformation module
#
# Contains routines used by routing protocol modules to adjust node/link BFD state
#
import typing

from box import Box

from . import _Module

class BFD(_Module):
  pass

def bfd_link_state(node: Box,proto: str) -> None:
  if not proto in node:           # pragma: no cover (impossible to be called from an IGP module if that module is not enabled)
    return
  if not 'bfd' in node[proto]:
    return
  if not 'interfaces' in node:    # pragma: no cover (pretty hard to get here ;)
    return

  node[proto].bfd = True if node[proto].bfd else False   # Convert protocol-level BFD setting into Boolean

  for l in node.interfaces:
    if not proto in l:                     # No protocol-specific link parameters?
      l[proto] = {}                        # ... start with an empty dictionary

    if not 'bfd' in l[proto]:              # No BFD protocol-specific parameters?
      l[proto].bfd = node[proto].bfd       # ... copy from node value and move on

    p = l[proto]
    disable_bfd = False
    disable_bfd = disable_bfd or ('bfd' in l and not l.bfd)     # BFD is disabled on the interface
    disable_bfd = disable_bfd or ('bfd' in p and not p.bfd)
    if disable_bfd:
      l[proto].pop('bfd',None)

    if not l[proto]:
      l.pop(proto,None)

def multiprotocol_bfd_link_state(node: Box,proto: str) -> None:
  if not proto in node:           # pragma: no cover (impossible to be called from an IGP module if that module is not enabled)
    return
  if not 'bfd' in node[proto]:
    return

  # Transform boolean BFD routing protocol state into per-AF-state
  #
  if isinstance(node[proto].bfd,bool):
    if node[proto].bfd:
      node[proto].bfd = {} 
      for af in ('ipv4','ipv6'):
        if af in node[proto].af:
          node[proto].bfd[af] = True
    else:
      node[proto].bfd = {}

  if not 'interfaces' in node:
    return

  for l in node.interfaces:
    if not proto in l:                     # No protocol-specific link parameters?
      l[proto] = {}                        # ... start with an empty dictionary

    if not 'bfd' in l[proto]:              # No BFD protocol-specific parameters?
      l[proto].bfd = node[proto].bfd       # ... copy from node value and move on

    p = l[proto]                           # Check whether BFD is disabled
    disable_bfd = False
    disable_bfd = disable_bfd or ('bfd' in l and not l.bfd)  # either on the interface
    disable_bfd = disable_bfd or ('bfd' in p and not p.bfd)  # ... or in protocol setting

    if disable_bfd:                        # If BFD is disabled
      l[proto].pop('bfd',None)             # ... remove all BFD protocol parameters
    else:
      if isinstance(l[proto].bfd,bool):    # Otherwise if we had 'bfd: True'
        l[proto].bfd = node[proto].bfd     # ... then copy node data into interface data

    if 'bfd' in l[proto]:                  # AF cleanup
      for af in ('ipv4','ipv6'):
        if not af in l:                    # Remove AF from proto.bfd if it's not enabled on the interface
          l[proto].bfd.pop(af,'None')
      if not l[proto].bfd:                 # If there's nothing left...
        l[proto].pop('bfd',None)           # ... remove proto.bfd dictionary

    if not l[proto]:                       # Last check: if we didn't get any useful protocol data
      l.pop(proto,None)                    # ... remove protocol data from the interface



#
# OSPF transformation module
#
import typing

from box import Box

from . import _Module,_routing
from . import bfd
from .. import common
from ..augment import devices

class OSPF(_Module):

  def node_post_transform(self, node: Box, topology: Box) -> None:
    features = devices.get_device_features(node,topology.defaults)

    _routing.router_id(node,'ospf',topology.pools)
    for intf in node.get('interfaces',[]):
      if not _routing.external(intf,'ospf'):
        _routing.passive(intf,'ospf')
        err = _routing.network_type(intf,'ospf',['point-to-point','point-to-multipoint','broadcast','non-broadcast'])
        if err:
          common.error(f'{err}\n... node {node.name} link {intf}')

    # We need to set ospf.unnumbered if we happen to have OSPF running over an unnumbered
    # link -- Arista EOS needs an extra nerd knob to make it work
    #
    # An interface can be unnumbered if it has the 'unnumbered' flag set or if it
    # has IPv4 enabled but no IPv4 address (ipv4: true)
    #
    # While doing that, we also check for the number of neighbors on unnumbered interfaces
    # and report an error if there are none (in which case the interface should not be unnumbered)
    # or more than one (in which case OSPF won't work)
    #
    for l in node.get('interfaces',[]):
      is_unnumbered = \
        'unnumbered' in l or \
        ('ipv4' in l and isinstance(l.ipv4,bool) and l.ipv4)
      if is_unnumbered and 'ospf' in l:
        node.ospf.unnumbered = True
        if len(l.get('neighbors',[])) > 1:
          common.error(
            f'OSPF does not work over multi-access unnumbered IPv4 interfaces: node {node.name} link {l.name}',
            common.IncorrectValue,
            'ospf')
        elif not len(l.get('neighbors',[])):
          common.error(
            f'Configuring OSPF on an unnumbered stub interface makes no sense: node {node.name} link {l.name}',
            common.IncorrectValue,
            'ospf')

    if 'unnumbered' in node.ospf:
      if not features.ospf.unnumbered:
        common.error(
          f'Device {node.device} used on node {node.name} cannot run OSPF over unnumbered interface',
          common.IncorrectValue,
          'interfaces')

    #
    # Final steps:
    # * move OSPF-enabled VRF interfaces into VRF dictionary
    # * Calculate address families
    # * Enable BFD
    # * Remove OSPF module if there are no OSPF-enabled global or VRF interfaces
    #
    _routing.build_vrf_interface_list(node,'ospf',topology)
    _routing.routing_af(node,'ospf')
    bfd.bfd_link_state(node,'ospf')
    _routing.remove_unused_igp(node,'ospf')



#
# netlab config command
#
# Deploy custom configuration template to network devices
#
import typing
import argparse
import os
import glob
from box import Box

from .. import common
from .. import read_topology

def print_table(heading: typing.List[str],rows: typing.List[typing.List[str]]) -> None:

  col_len: typing.List[int] = []

  def print_row(separator: str, row: typing.Optional[list] = None, char: str = ' ') -> None:
    line = separator
    for idx,clen in enumerate(col_len):
      if row:
        value = ' ' + row[idx] + (' ' * 80)
      else:
        value = char * (clen + 2)
      line = line + value[:clen+2] + separator
    print(line)

  for idx,data in enumerate(heading):
    slice_len = [ len(k[idx]) for k in rows ]
    slice_len.append(len(heading[idx]))
    col_len.append(max(slice_len))

  print_row('+',char='-')
  print_row('|',row=heading)
  print_row('+',char='=')
  for idx,row in enumerate(rows):
    print_row('|',row=row)
    print_row('+',char='-')

def show_images(settings: Box, args: argparse.Namespace) -> None:
  heading = ['device']
  heading.extend(settings.providers.keys())

  rows = []
  for device in settings.devices.keys():
    if device == 'none':
      continue

    if args.device and device != args.device:
      continue

    row = [ device ]
    for p in heading[1:]:
      row.append(settings.devices[device][p].get("image",""))
    rows.append(row)

  print((args.device or "Device") + " image names by virtualization provider")
  print("")
  print_table(heading,rows)

def show_module_support(settings: Box, args: argparse.Namespace) -> None:
  heading = ['device']
  heading.extend([ m for m in settings.keys() if 'supported_on' in settings[m]])

  rows = []
  for device in sorted(settings.devices.keys()):
    if device == 'none':
      continue

    if args.device and device != args.device:
      continue

    row = [ device ]
    for m in heading[1:]:
      value = "x" if device in settings[m].supported_on else ""
      row.append(value)
    rows.append(row)

  print("Configuration modules supported by " + (args.device or "individual devices"))
  print("")
  print_table(heading,rows)

show_dispatch = {
  'images': show_images,
  'module-support': show_module_support
}

#
# CLI parser for 'netlab show' command
#
def show_parse(args: typing.List[str]) -> argparse.Namespace:
  global show_dispatch
  parser = argparse.ArgumentParser(
    prog='netlab show',
    description='Display system settings')
  parser.add_argument(
    '-d','--device',
    dest='device',
    action='store',
    help='Display system information for a single device')
  parser.add_argument(
    dest='action',
    action='store',
    choices=show_dispatch.keys(),
    help='Select the system information to display')

  return parser.parse_args(args)

def run(cli_args: typing.List[str]) -> None:
  global show_dispatch
  args = show_parse(cli_args)
  settings =  read_topology.read_yaml("package:topology-defaults.yml")

  if settings is None:
    common.fatal("Cannot read system settings")
  else:
    show_dispatch[args.action](settings,args)



#
# A no-op output module, useful when testing new features
#
import typing

from box import Box

from .. import common
from . import _TopologyOutput

class NONE(_TopologyOutput):

  def write(self, topo: Box) -> None:
    pass



#
# Generic data model manipulation routines
#

import typing
from box import Box
from .. import common

#
# Change all NULL values in a nested dictionary structure to empty strings
# to make them nicer in YAML printouts
#
def null_to_string(d: typing.Dict) -> None:
  for k in d.keys():
    if isinstance(d[k],dict):
      null_to_string(d[k])
    elif d[k] is None:
      d[k] = ""

#
# Safe get from a hierarchical dictionary (won't create new objects)
#

def get_from_box(b: Box, selector: typing.Union[str,typing.List[str]], partial: bool = False) -> typing.Optional[typing.Any]:
  if isinstance(selector,str):
    selector = selector.split('.')

  for idx,k in enumerate(selector):
    if not k in b:
      return b if partial and idx > 0 else None   # return partial result if request assuming we got at least one match before

    if not isinstance(b[k],dict):                                       # we are at a leaf node
      return b[k] if partial or idx == len(selector) - 1 else None      # ... return the value if we're at the end of
                                                                        # ... the chain or accept partial lookup
    b = b[k]

  return b

#
# Set a dictionary value specified by a list of keys
#
def set_dots(b : dict,k_list : list,v : typing.Any) -> None:
  if len(k_list) <= 1:
    b[k_list[0]] = v
    return
  if not k_list[0] in b:
    b[k_list[0]] = {}
  elif b[k_list[0]] is None:
    b[k_list[0]] = {}
  set_dots(b[k_list[0]],k_list[1:],v)

#
# Change dotted dictionary keys into nested dictionaries
#
def unroll_dots(b : typing.Any) -> None:
  if isinstance(b,dict):
    for k in list(b.keys()):
      unroll_dots(b[k])
      if isinstance(k,str) and ('.' in k):
        v = b[k]
        del b[k]     # If you're using Box with box_dots parameter
        set_dots(b,k.split('.'),v)
  elif isinstance(b,list):
    for v in b:
      unroll_dots(v)
  else:
    return

#
# must_be_list: make sure a dictionary value is a list. Convert scalar values
#   to list if needed, report an error otherwise.
#
# must_be_string: make sure a dictionary value is a string. Throw an error otherwise.
#
# Input arguments:
#   parent - the parent dictionary of the attribute we want to listify
#            (a pointer to the element would be even better, but Python)
#   key    - the parent dictionary key
#   path   - path of the parent dictionary that would help the user identify
#            where the problem is
#
# Sample use: make sure the 'config' attribute of a node is list
#
#    must_be_list(node,'config',f'nodes.{node.name}')
#
def must_be_list(parent: Box, key: str, path: str) -> typing.Optional[list]:
  if not key in parent:
    parent[key] = []
    return parent[key]

  if isinstance(parent[key],list):
    return parent[key]

  if isinstance(parent[key],(str,int,float,bool)):
    parent[key] = [ parent[key] ]
    return parent[key]

  wrong_type = "dictionary" if isinstance(parent[key],dict) else str(type(parent[key]))
  common.error(
  	f'attribute {path}.{key} must be a scalar or a list, found {wrong_type}',
  	common.IncorrectType)
  return None

def must_be_string(parent: Box, key: str, path: str) -> typing.Optional[str]:
  if not key in parent:
    return None

  if isinstance(parent[key],str):
    return parent[key]

  wrong_type = "dictionary" if isinstance(parent[key],dict) else str(type(parent[key]))
  common.error(
    f'attribute {path}.{key} must be a string, found {wrong_type}',
    common.IncorrectType)
  return None

#
# must_be_bool: check whether a parameter is a boolean value and remove False value
#   to simplify Jinja2 templates
#
# Input arguments:
#   parent - parent dictionary
#   key    - parent dictionary key
#   path   - parent dictionary path (used in error messages)
#
def must_be_bool(parent: Box, key: str, path: str) -> None:
  if not key in parent:
    return

  if isinstance(parent[key],bool):
    if not parent[key]:
      parent.pop(key,None)
    return

  common.error(
    f'attribute {path}.{key} must be a boolean, found {str(type(parent[key]))}',
    common.IncorrectType)

"""
bool_to_defaults: 

* remove a parameter set to False just to prevent default propagation
* replace a True value with a default dictionary
* dive recursively into all keys specified in default dictionary
"""

def bool_to_defaults(obj: Box, attr: str, defvalue: typing.Optional[typing.Any] = None) -> None:
  if not attr in obj:                   # Attribute not in dictionary, nothing to do
    return

  if isinstance(obj[attr],Box):
    for k in list(obj[attr].keys()):
      if isinstance(defvalue,dict) and k in defvalue:
        bool_to_defaults(obj[attr],k,defvalue[k])
    return

  if not isinstance(obj[attr],bool):    # Attribute not a boolean, no further work needed
    return

  if not obj[attr]:                     # Remove False value
    obj.pop(attr,None)
    return

  if not defvalue is None:              # If the default value was specified, replace True with default value
    obj[attr] = defvalue

"""
validate_list_elements: check whether the elements of a list belong to a set of predefined values
"""

def validate_list_elements(data: list, values: list, path: str) -> bool:
  if not isinstance(data,list):  # pragma: no cover
    common.fatal(f'{path} should be a list')
    return False
  return all(item in values for item in data)



#
# netlab config command
#
# Deploy custom configuration template to network devices
#
import typing
import argparse
import os
import string
import pathlib
import glob
import subprocess
import shutil

from box import Box

from .. import common
from .. import read_topology
from . import external_commands
from . import collect
from . import fs_cleanup

def tarball_parse(args: typing.List[str], settings: Box) -> argparse.Namespace:
  parser = argparse.ArgumentParser(
    prog='netlab clab tarball',
    description='Create a ready-to-use tarball containing containerlab configuration file and startup configs')
  parser.add_argument(
    '-v','--verbose',
    dest='verbose',
    action='count',
    default=0,
    help='Verbose logging')
  parser.add_argument(
    '-q','--quiet',
    dest='quiet',
    action='store_true',
    help='Run Ansible playbook and tar with minimum output')
  parser.add_argument(
    '--config',
    dest='output',
    action='store',
    nargs='?',
    default='config',
    help='Startup configuration directory (default: config)')
  parser.add_argument(
    '--cleanup',
    dest='cleanup',
    action='store_true',
    help='Clean up config directory and modified configuration file after creating tarball')
  parser.add_argument(
    dest='tarball',
    action='store',
    type=argparse.FileType('w'),
    help='Destination tarball (.tar.gz will be added if needed)')
  return parser.parse_args(args)

def find_config_file(n: str, cfglist: typing.List[str]) -> typing.Optional[str]:
  for fname in cfglist:
    if fname == n:
      return fname
    if fname.find(n+'.') == 0:
      return fname

  return None

def clab_config_adjust(infile: str, outfile: str, configs: str) -> None:
  clab = read_topology.read_yaml(infile)
  if not clab:
    common.fatal("Cannot read clab.yml configuration file, aborting")
    return

  clab_yml = clab.to_yaml()
  if not ('topology' in clab and 'nodes' in clab.topology):
    common.fatal('Containerlab configuration file {infile} is weird: cannot find topology.nodes dictionary')

  try:
    cfglist = os.listdir(configs)
  except Exception as ex:
    common.fatal('Cannot read the contents of {configs} directory: {ex}')

  for n in list(clab.topology.nodes.keys()):
    cfgfile = find_config_file(n,cfglist)
    if cfgfile:
      cfgfile = f"{configs}/{cfgfile}"
      print(f"Found config file for {n}: {cfgfile}")
      clab.topology.nodes[n]['startup-config'] = cfgfile

  if clab.to_yaml() == clab_yml:
    common.fatal('No relevant configuration files were found in {configs} directory, aborting')

  output = common.open_output_file(outfile)
  output.write(clab.to_yaml())
  output.close()

def clab_tarball(cli_args: typing.List[str], settings: Box) -> None:
  args = tarball_parse(cli_args,settings)

  if not os.path.exists('clab.yml'):
    common.fatal('Containerlab configuration file clab.yml not found, aborting...')

  external_commands.print_step(1,"Collecting device configurations")
  os.environ["ANSIBLE_STDOUT_CALLBACK"] = "dense"
  collect.run(['-o',args.output])

  external_commands.print_step(2,"Adjusting containerlab configuration file",spacing = True)
  clab_config_adjust(infile='clab.yml',outfile='clab.config.yml',configs=args.output)

  external_commands.print_step(3,"Creating tarball",spacing = True)
  tarball = collect.get_tarball_file(args.tarball.name)

  try:
    subprocess.check_call(['tar','cfz' if args.quiet else 'cvfz',tarball,'clab.config.yml',args.output])
  except Exception as ex:
    common.fatal(f"Cannot start tar: {ex}")

  if args.cleanup:
    external_commands.print_step(4,"Cleanup",spacing = True)
    fs_cleanup([ args.output,'clab.config.yml' ])
    print("... done")

def clab_usage() -> None:
  print("Usage: netlab clab tarball --help")

def run(cli_args: typing.List[str]) -> None:
  settings = read_topology.read_yaml('package:topology-defaults.yml')
  if not cli_args:
    clab_usage()
    return

  if not settings:
    common.fatal("Cannot read the system defaults","libvirt")
    return

  if cli_args[0] == 'tarball':
    clab_tarball(cli_args[1:],settings)
  else:
    clab_usage()



#
# Run external commands from netlab CLI
#
import typing
import argparse
import os
import glob
import subprocess
import shutil

from box import Box
from pathlib import Path

from .. import common

def print_step(n: int, txt: str, spacing: typing.Optional[bool] = False) -> None:
  if spacing:
    print()
  print("Step %d: %s" % (n,txt))
  print("=" * 60)

def stringify(cmd : typing.Union[str,list]) -> str:
  if isinstance(cmd,list):
    return " ".join(cmd)
  return str(cmd)

def run_command(cmd : typing.Union[str,list], check_result : bool = False) -> bool:
  if common.DEBUG:
    print("Not running: %s" % cmd)
    return True

  if common.VERBOSE:
    print(".. executing: %s" % cmd)

  if isinstance(cmd,str):
    cmd = [ arg for arg in cmd.split(" ") if arg not in (""," ") ]

  try:
    result = subprocess.run(cmd,capture_output=check_result,check=True,text=True)
    if not check_result:
      return True
    return result.stdout != ""
  except Exception as ex:
    if not common.QUIET:
      print( f"Error executing {stringify(cmd)}:\n  {ex}" )
    return False

def test_probe(p : str) -> bool:
  return run_command(p,True)

def set_ansible_flags(cmd : list) -> list:
  if common.VERBOSE:
    cmd.append("-" + "v" * common.VERBOSE)

  if common.QUIET:
    os.environ["ANSIBLE_STDOUT_CALLBACK"] = "selective"

  return cmd

def run_probes(settings: Box, provider: str, step: int = 0) -> None:
  if step:
    print_step(step,"Checking virtualization provider installation",spacing = True)
  elif common.VERBOSE:
    print("Checking virtualization provider installation")
  for p in settings.providers[provider].probe:
    if not test_probe(p):
      common.fatal("%s failed, aborting" % p)
  if common.VERBOSE or step:
    print(".. all tests succeeded, moving on\n")

def start_lab(settings: Box, provider: str, step: int = 2, command: str = "test") -> None:
  print_step(step,"starting the lab")
  cmd = settings.providers[provider].start
  if not run_command(cmd):
    common.fatal("%s failed, aborting..." % cmd,command)

def deploy_configs(step : int = 3, command: str = "test", fast: typing.Optional[bool] = False) -> None:
  print_step(step,"deploying initial device configurations",spacing = True)
  cmd = ["netlab","initial"]
  if common.VERBOSE:
    cmd.append("-" + "v" * common.VERBOSE)

  if os.environ.get('NETSIM_FAST_CONFIG',None) or fast:
    cmd.append("--fast")

  if not run_command(set_ansible_flags(cmd)):
    common.fatal("netlab initial failed, aborting...",command)

def custom_configs(config : str, group: str, step : int = 4, command: str = "test") -> None:
  print_step(step,"deploying custom configuration template %s for group %s" % (config,group))
  cmd = ["netlab","config",config,"--limit",group]

  if not run_command(set_ansible_flags(cmd)):
    common.fatal("netlab config failed, aborting...",command)

def stop_lab(settings: Box, provider: str, step: int = 4, command: str = "test") -> None:
  print_step(step,"stopping the lab",True)
  cmd = settings.providers[provider].stop
  if not run_command(cmd):
    common.fatal("%s failed, aborting..." % cmd,command)



'''
Create detailed node-level data structures from topology

* Discover desired imagex (boxes)
* Add default module list to nodes without specific modules
* Set loopback and management interface data
'''
import typing

from box import Box

from .. import common
from .. import addressing
from . import devices

#
# Starting with release 1.1, the nodes data structure is a dictionary. Convert
# lists of dictionaries or lists of strings into a unified dictionary structure
#
def create_node_dict(nodes: Box) -> Box:
  if isinstance(nodes,dict):
    node_dict = nodes
  else:
    node_dict = Box({},default_box=True,box_dots=True)
    node_id = 0
    for n in nodes or []:
      if isinstance(n,dict):
        if not 'name' in n:
          common.error(f'Node is missing a "name" attribute: {n}',common.IncorrectValue,'nodes')
          continue
      elif isinstance(n,str):
        n = Box({ 'name': n },default_box=True,box_dots=True)
      node_id = node_id + 1
      n.id = node_id
      node_dict[n.name] = n

  for name in list(node_dict.keys()):
    ndata = node_dict[name]
    if ndata is None:
      ndata = Box({'name': name},default_box=True)
    elif not isinstance(ndata,dict):
      common.error(f'Node data for node {name} must be a dictionary')
      node_dict[name] = { 'name': name, 'extra': ndata }
      ndata = node_dict[name]
    else:
      ndata['name'] = name

    ndata.interfaces = ndata.interfaces or []   # Make sure node.interfaces is always defined
    node_dict[name] = ndata

  common.exit_on_error()
  return node_dict

def augment_mgmt_if(node: Box, defaults: Box, addrs: typing.Optional[Box]) -> None:
  if 'ifname' not in node.mgmt:
    mgmt_if = devices.get_device_attribute(node,'mgmt_if',defaults)
    if not mgmt_if:
      ifname_format = devices.get_device_attribute(node,'interface_name',defaults)
      if not isinstance(ifname_format,str):
        common.fatal("Missing interface name template for device type %s" % node.device)
        return

      ifindex_offset = devices.get_device_attribute(node,'ifindex_offset',defaults)
      if ifindex_offset is None:
        ifindex_offset = 1
      mgmt_if = ifname_format % (ifindex_offset - 1)
    node.mgmt.ifname = mgmt_if

  if addrs:
    for af in 'ipv4','ipv6':
      pfx = af + '_pfx'
      if pfx in addrs:
        if not addrs.get('start'):
          common.fatal("Start offset missing in management address pool for AF %s" % af)
        if not af in node.mgmt:
          node.mgmt[af] = str(addrs[pfx][node.id+addrs.start])

    if addrs.mac_eui and not 'mac' in node.mgmt:
      addrs.mac_eui[5] = node.id
      node.mgmt.mac = str(addrs.mac_eui)

"""
Add provider data to nodes:

* Check whether the node device exists
* Copy device.provider.node into node.provider
* Get device image
"""
def augment_node_provider_data(topology: Box) -> None:
  if not topology.defaults.devices:
    common.fatal('Device defaults (defaults.devices) are missing')

  for name,n in topology.nodes.items():
    if 'device' not in n:
      n.device = topology.defaults.device

    if not n.device:
      common.error(
        f'No device type specified for node {name} and there is no default device type',
        common.MissingValue,
        'nodes')
      continue

    devtype = n.device

    if not devtype in topology.defaults.devices:
      common.error(f'Unknown device {devtype} in node {name}',common.IncorrectValue,'nodes')
      continue

    if not isinstance(topology.defaults.devices[devtype],dict):
      common.fatal(f"Device data for device {devtype} must be a dictionary")

    provider = devices.get_provider(n,topology.defaults)
    pdata = devices.get_provider_data(n,topology.defaults)
    if 'node' in pdata:
      if not isinstance(pdata.node,Box):    # pragma: no cover
        common.fatal(f"Node data for device {devtype} provider {provider} must be a dictionary")
        return
      n[provider] = pdata.node + n.get(provider,{})

    if n.box:
      continue
    if 'image' in n:
      n.box = n.image
      del n['image']
      continue

    if 'image' in topology.defaults.devices[devtype]:
      image = topology.defaults.devices[devtype].image
      if isinstance(image,str):
        n.box = image
        continue
      else:
        common.error(
          f"Image attribute of device {devtype} used by node {name} should be a string\n... found {image}",
          common.IncorrectValue,
          'nodes')
        continue

    if 'image' in pdata:
      if isinstance(pdata.image,str):
        n.box = pdata.image
        continue
      else:
        common.error(
          f"Image attribute specified for provider {provider} on device {devtype} should be a string\n... found {pdata.image}",
          common.MissingValue,
          'nodes')
        continue

    common.error(
      f'No image specified for device {devtype} (provider {provider}) used by node {name}',
      common.MissingValue,
      'nodes')

"""
Add system data to devices -- hacks that are not yet covered in the settings structure
"""
def augment_node_system_data(topology: Box) -> None:
  if 'mtu' in topology.defaults.get('interfaces',{}):
    if not isinstance(topology.defaults.interfaces.mtu,int):            # pragma: no cover
      common.error(
        'defaults.interfaces.mtu setting should be an integer',
        common.IncorrectValue,
        'topology')
    else:
      for n in topology.nodes.values():
        if not 'mtu' in n:
          n.mtu = topology.defaults.interfaces.mtu
        else:
          if not isinstance(n.mtu,int):                                 # pragma: no cover
            common.error(
              f'nodes.{n.name}.mtu setting should be an integer',
              common.IncorrectValue,
              'nodes')

'''
get_next_id: given a list of static IDs and the last ID, get the next device ID
'''
def get_next_id(id_list: list, id: int) -> int:
  while id < 254:
    id = id + 1
    if not id in id_list:
      return id

  common.fatal(
    'Cannot get the next device ID. The lab topology is probably too big')  # pragma: no cover (I'm not going to write a test case for this one)
  return -1                                                                 # pragma: no cover (making mypy happy)

"""
augment_node_device_data: copy attributes that happen to be node attributes from device defaults into node data
"""

def augment_node_device_data(n: Box, defaults: Box) -> None:
  node_attr = defaults.attributes.get('node',[])
  dev_data  = devices.get_consolidated_device_data(n,defaults)

  for attr in node_attr:                      # Copy known node attributes from device+provider data into node data
    if attr in dev_data and not attr in n:
      n[attr] = dev_data[attr]

  if 'node' in defaults.devices[n.device]:    # Copy everything within device (but not provider) node dictionary into node data
    for k in defaults.devices[n.device].node.keys():
      if not k in n:
        n[k] = defaults.devices[n.device].node[k]

'''
Main node transformation code

* set node ID
* set loopback address(es)
* copy device data from defaults
* set management IP and MAC addresses
'''
def transform(topology: Box, defaults: Box, pools: Box) -> None:
  id = 0
  id_list = []

  for name,n in topology.nodes.items():
    if 'id' in n:
      if isinstance(n.id,int) and n.id > 0 and n.id <= 250:
        id_list.append(n.id)
      else:
        common.error('Device ID must be an integer between 1 and 250')

  common.exit_on_error()


  for name,n in topology.nodes.items():
    if not 'id' in n:
      id = get_next_id(id_list,id)
      n.id = id

    if not n.name: # pragma: no cover (name should have been checked way before)
      common.fatal(f"Internal error: node does not have a name {n}",'nodes')
      return

    augment_node_device_data(n,defaults)

    if pools.loopback and n.get('role','') != 'host':
      prefix_list = addressing.get(pools,['loopback'],n.id)
      for af in prefix_list:
        if not n.loopback[af]:
          if af == 'ipv6':
            n.loopback[af] = addressing.get_addr_mask(prefix_list[af],1)
          else:
            n.loopback[af] = str(prefix_list[af])

    augment_mgmt_if(n,defaults,topology.addressing.mgmt)
    topology.Provider.call("augment_node_data",n,topology)



#
# Create Ansible inventory
#
import typing

import yaml
import os
import sys
from box import Box

from .. import common
from . import _TopologyOutput
from .common import adjust_inventory_host

def create(nodes: Box, defaults: Box, addressing: typing.Optional[Box] = None) -> Box:
  inventory = Box({},default_box=True,box_dots=True)

  if addressing:
    inventory.all.vars.pools = addressing
    for name,pool in inventory.all.vars.pools.items():
      for k in list(pool.keys()):
        if ('_pfx' in k) or ('_eui' in k):
          del pool[k]

  for name,node in nodes.items():
    inventory[name] = adjust_inventory_host(
                        node = node,
                        defaults = defaults,
                        ignore = ['name'],
                        group_vars = True)

  return inventory

def write_yaml(data: Box, fname: str, header: str) -> None:
  dirname = os.path.dirname(fname)
  if dirname and not os.path.exists(dirname):
    os.makedirs(dirname)

  with open(fname,"w") as output:
    output.write(header)
    if callable(getattr(data,"to_yaml",None)):
      output.write(data.to_yaml())
    else:                            # pragma: no cover -- this should never happen as we're using Box, but just in case...
      output.write(yaml.dump(data))
    output.close()

min_inventory_data = [ 'id','ansible_host','ansible_port' ]

def write_devices(data: Box, fname: str, fmt: typing.Optional[str]) -> None:
  fmt = fmt or 'standard'
  addressing = data.get('addressing',{}) if 'pools' in fmt else None

  inventory = create(data.nodes,data.get('defaults',{}),addressing)

  header = "# Netsim Devices inventory created from %s\n#\n---\n" % data.get('input','<unknown>')

  write_yaml(inventory,fname,header)
  print("Created Netsim Devices inventory %s" % fname)

def read_inventory(host: str, filename: typing.Optional[str] = None) -> typing.Optional[dict]:
  filename = filename or os.getenv('NETSIM_DEVICES') or 'netsim-devices.yml'
  if not os.path.isfile(filename):
    common.fatal('Cannot read netsim device inventory: %s does not exist')
    return None
  try:
    data = Box().from_yaml(filename=filename,default_box=True,box_dots=True,default_box_none_transform=False)
  except:
    common.fatal("Cannot read YAML from %s: %s " % (filename,str(sys.exc_info()[1])))

  if host in (data or {}):
    return data[host]

  return None

class DeviceInventory(_TopologyOutput):

  def write(self, topology: Box) -> None:
    hostfile = self.settings.hostfile or 'netsim-devices.yml'
    output_format = None

    if hasattr(self,'filenames'):
      hostfile = self.filenames[0]
      if len(self.filenames) > 1:
        common.error('Extra output filename(s) ignored: %s' % str(self.filenames[1:]),common.IncorrectValue,'devices')

    if self.format:
      output_format = self.format[0]

    write_devices(topology,hostfile,output_format)



#
# Common inventory-related routines (used by Ansible and Devices)
#

import typing
from box import Box
from ..augment import devices

topo_to_host = { 'mgmt.ipv4': 'ansible_host', 'hostname': 'ansible_host', 'id': 'id' }
topo_to_host_skip = [ 'name','device' ]

def provider_inventory_settings(node: Box, defaults: Box) -> None:
  p_data = defaults.providers[defaults.provider]
  if not p_data:
    return        # pragma: no cover -- won't create an extra test case just to cover the "do nothing" scenario

  if 'inventory' in p_data:
    for k,v in p_data['inventory'].items():
      node[k] = v

  if 'inventory_port_map' in p_data and 'forwarded' in p_data:
    for k,v in p_data['inventory_port_map'].items():
      if k in p_data['forwarded']:
        node[v] = p_data['forwarded'][k] + node['id']

def add_group_vars(
      host: Box,
      node: Box,
      defaults: Box) -> None:

  group_vars = devices.get_device_attribute(node,'group_vars',defaults)
  if isinstance(group_vars,dict):
    for (k,v) in group_vars.items():
      host[k] = v

def adjust_inventory_host(
      node: Box,
      defaults: Box,
      translate: typing.Optional[dict] = None,
      ignore: typing.Optional[list] = None,
      group_vars: typing.Optional[bool] = False) -> Box:
  host = Box({})

  translate = translate or topo_to_host
  ignore = ignore or topo_to_host_skip

  if group_vars:
    add_group_vars(host,node,defaults)

  for (node_key,inv_key) in translate.items():
    if "." in node_key:
      value = node[node_key]
    else:
      value = node.get(node_key,None)
    if value:
      host[inv_key] = value

  for (k,v) in node.items():
    if not k in ignore:
      host[k] = v

  provider_inventory_settings(host,defaults)
  return host




#
# Plugin API routines
#
import sys
import typing
import os

from box import Box

from . import common
from . import data

def get_config_name(g: dict) -> typing.Optional[str]:
  config_name = g.get('config_name',None)
  if config_name:
    return config_name

  common.fatal("Cannot get configuration template name for plugin %s" % g.get('__file__'),'plugin')
  return None

def node_config(node: Box, config_name: typing.Optional[str]) -> None:
  if config_name:
    config = node.get('config',[])
    if not config_name in config:
      node.config =  config + [ config_name ]

def list_attribute(parent: Box, key: str, path: str) -> typing.Optional[list]:
  return data.must_be_list(parent,key,path)



"""
Generic dynamic loadable module framework

The Callback class defines two methods:

* load: tries to load a more specific module
* call: tries to call named method with specified parameters
"""

import importlib
import inspect
import typing

from . import common

class Callback():

  @classmethod
  def find_class(self, module_name: str) -> typing.Optional[typing.Any]:
    if common.VERBOSE:
      print("loading %s..." % module_name)
    try:
      module = importlib.import_module(module_name)
      for name,obj in inspect.getmembers(module):
        if inspect.isclass(obj) and issubclass(obj,Callback):
          if common.VERBOSE:
            print("... found %s " % obj)
          return obj
      return None

    except (ImportError, AttributeError):
      if common.VERBOSE:
        print("Failed to load specific module")
      return None

  def call(self, name: str, *args: typing.Any, **kwargs: typing.Any) -> None:
    method = getattr(self,name,None)
    if method:
      method(*args, **kwargs)



#
# netlab config command
#
# Deploy custom configuration template to network devices
#
import typing
import argparse
import os
import glob
import subprocess
from pathlib import Path

from .. import common

#
# CLI parser for 'netlab install' command
#
def install_parse(args: typing.List[str]) -> argparse.Namespace:
  moddir = Path(__file__).resolve().parent.parent
  choices = map(
    lambda x: Path(x).stem,
    glob.glob(str(moddir / "install/*sh")))

  parser = argparse.ArgumentParser(
    prog='netlab install',
    description='Install additional software')
  parser.add_argument(
    '-v','--verbose',
    dest='verbose',
    action='store_true',
    help='Verbose logging')
  parser.add_argument(
    '-q','--quiet',
    dest='quiet',
    action='store_true',
    help='Be as quiet as possible')
  parser.add_argument(
    '-y','--yes',
    dest='yes',
    action='store_true',
    help='Run the script without prompting for a confirmation')
  parser.add_argument(
    dest='script',
    action='store',
    choices=list(choices),
    nargs="*",
    help='Run the specified installation script')

  return parser.parse_args(args)

def run(cli_args: typing.List[str]) -> None:
  if not cli_args:
    common.fatal("Specify an installation script to run or use -h to get help","install")
    return

  args = install_parse(cli_args)

  moddir = Path(__file__).resolve().parent.parent
  env = { key:value for (key,value) in os.environ.items() }
  if args.verbose:
    env['FLAG_APT'] = '-V'
    env['FLAG_PIP'] = '-v'

  if args.quiet:
    env['FLAG_APT'] = '-qq'
    env['FLAG_QUIET'] = '-qq'
    env['FLAG_PIP'] = '-qq'

  if args.yes:
    env['FLAG_YES'] = 'Y'

  for script in args.script:
    script_path = str(moddir / "install" / (script+".sh"))
    if args.verbose:
      print("Running installation script: %s" % script_path)

    try:
      subprocess.run(
        ['bash',script_path],
        check=True,
        env=env)
    except:
      print("\nInstallation script %s.sh failed, exiting..." % script)
      return



#
# IS-IS transformation module
#
from box import Box

from . import _Module,_routing

class EIGRP(_Module):

  def node_post_transform(self, node: Box, topology: Box) -> None:
    _routing.router_id(node,'eigrp',topology.pools)
    for intf in node.get('interfaces',[]):
      if not _routing.external(intf,'eigrp'):
        _routing.passive(intf,'eigrp')

    _routing.routing_af(node,'eigrp')



#
# BGP transformation module
#
import typing

import re
from box import Box
import netaddr

from . import _Module,_routing
from .. import common
from .. import data
from ..augment.links import IFATTR

def check_bgp_parameters(node: Box) -> None:
  if not "bgp" in node:  # pragma: no cover (should have been tested and reported by the caller)
    return
  if not "as" in node.bgp or not isinstance(node.bgp.get('as',{}),int):
    common.error("Node %s has BGP enabled but no AS number specified" % node.name)

  if "community" in node.bgp:
    bgp_comm = node.bgp.community
    if isinstance(bgp_comm,str):
      node.bgp.community = { 'ibgp' : [ bgp_comm ], 'ebgp': [ bgp_comm ]}
    elif isinstance(bgp_comm,list):
      node.bgp.community = { 'ibgp' : bgp_comm, 'ebgp': bgp_comm }
    elif not(isinstance(bgp_comm,dict)):
      common.error("bgp.community attribute in node %s should be a string, a list, or a dictionary (%s)" % (node.name,str(bgp_comm)))
      return

    for k in node.bgp.community.keys():
      if not k in ['ibgp','ebgp']:
        common.error("Invalid BGP community setting in node %s: %s" % (node.name,k))
      if isinstance(node.bgp.community[k],str):
        node.bgp.community[k] = [ node.bgp.community[k] ]
      for v in node.bgp.community[k]:
        if not v in ['standard','extended']:
          common.error("Invalid BGP community propagation setting for %s sessions in node %s: %s" % (k,node.name,v))

def find_bgp_rr(bgp_as: int, topology: Box) -> typing.List[Box]:
  return [ n 
    for n in topology.nodes.values() 
      if 'bgp' in n and n.bgp["as"] == bgp_as and n.bgp.get("rr",None) ]

def bgp_neighbor(n: Box, intf: Box, ctype: str, extra_data: typing.Optional[dict] = None) -> Box:
  ngb = Box(extra_data or {},default_box=True,box_dots=True)
  ngb.name = n.name
  ngb["as"] = n.bgp.get("as")
  ngb["type"] = ctype
  for af in ["ipv4","ipv6"]:
    if af in intf:
      if "unnumbered" in ngb and ngb.unnumbered == True:
        ngb[af] = True
      else:
        ngb[af] = str(netaddr.IPNetwork(intf[af]).ip)
  return ngb

def get_neighbor_rr(n: Box) -> typing.Optional[typing.Dict]:
  if "rr" in n.get("bgp"):
    return { "rr" : n.bgp.rr }

  return {}

class BGP(_Module):

  '''
  process_as_list:
    If the global BGP parameters have as_list attribute, set node AS numbers and node
    RR flags accordingly
  '''
  def process_as_list(self, topology: Box) -> None:
    if not 'bgp' in topology:                            # Do we have global BGP settings?
      return

    if not 'as_list' in topology.bgp:                    # Do we have global AS list?
      return

    if not isinstance(topology.bgp.as_list,dict):
      common.error(
        "bgp.as_list should be a dictionary of AS parameters",
        common.IncorrectValue)
      return

    node_data = Box({},default_box=True,box_dots=True)
    for asn,as_data in topology.bgp.as_list.items():
      if not isinstance(as_data,Box):
        common.error(
          "Invalid value in bgp.as_list for ASN %s: " % asn + \
          "\n... Each ASN in a BGP as_list must be a dictionary with (at least) members key:"+
          "\n... Found: %s" % as_data,
          common.IncorrectValue)
        continue

      if not 'members' in as_data:
        common.error(
          "BGP as_list for ASN %s does not have a member attribute" % asn,
          common.IncorrectValue)
        continue

      data.must_be_list(as_data,'members',f'bgp.as_list.{asn}')
      for n in as_data.members:
        if not n in topology.nodes:
          common.error(
            "Invalid node name %s in member list of BGP AS %s" % (n,asn),
            common.IncorrectValue)
          continue
        elif 'as' in node_data[n]:
          common.error(
            f"BGP module supports at most 1 AS per node; {n} is already member of {node_data[n]['as']} and cannot also be part of {asn}",
            common.IncorrectValue)
          continue
        node_data[n]["as"] = asn

      for n in as_data.get('rr',{}):
        if not n in topology.nodes:
          common.error(
            "Invalid node name %s in route reflector list of BGP AS %s" % (n,asn),
            common.IncorrectValue)
          continue
        if node_data[n]["as"] != asn:
          common.error(
            "Node %s is specified as route reflector in AS %s but is not in member list" % (n,asn),
            common.IncorrectValue)
          continue
        node_data[n].rr = True

    for name,node in topology.nodes.items():
      if name in node_data:
        node_as = node.bgp.get("as",None)
        if node_as and node_as != node_data[name]["as"]:
          common.error(
            "Node %s has AS %s but is also in member list of AS %s" % (node.name,node_as,node_data[node.name]["as"]),
            common.IncorrectValue)
          continue

        node.bgp = node_data[name] + node.bgp

  '''
  bgp_build_group: create automatic groups based on BGP AS numbers
  '''
  def build_bgp_groups(self, topology: Box) -> None:
    for gname,gdata in topology.groups.items():
      if re.match('as\\d+$',gname):
        if gdata.get('members',None):
          common.error('BGP AS groups should not have static members %s' % gname)

    for name,node in topology.nodes.items():
      if 'bgp' in node and 'as' in node.bgp:
        grpname = "as%s" % node.bgp["as"]
        if not grpname in topology.groups:
          topology.groups[grpname] = { 'members': [] }

        if not 'members' in topology.groups[grpname]:   # pragma: no cover (members list is created in group processing)
          topology.groups[grpname].members = []

        topology.groups[grpname].members.append(name)

  """
  Module pre-default:

  * process AS list
  * create automatic BGP groups
  """
  def module_pre_default(self, topology: Box) -> None:
    self.process_as_list(topology)
    self.build_bgp_groups(topology)

  """
  Node pre-transform: set bgp.rr node attribute to _true_ if the node name is in the
  global bgp.rr attribute. Also, delete the global bgp.rr attribute so it's not propagated
  down to nodes
  """
  def node_pre_transform(self, node: Box, topology: Box) -> None:
    if "rr_list" in topology.get("bgp",{}):
      if node.name in topology.bgp.rr_list:
        node.bgp.rr = True

  """
  Link pre-transform: Set link role based on BGP nodes attached to the link.

  If the nodes belong to at least two autonomous systems, and the ebgp_role
  variable is set, set the link role to ebgp_role
  """
  def link_pre_transform(self, link: Box, topology: Box) -> None:
    ebgp_role = topology.bgp.get("ebgp_role",None) or topology.defaults.bgp.get("ebgp_role",None)
    if not ebgp_role:
      return

    as_set = {}
    for ifdata in link.get(IFATTR,[]):
      n = ifdata.node
      if "bgp" in topology.nodes[n]:
        node_as = topology.nodes[n].bgp.get("as")
        if node_as:
          as_set[node_as] = True

    if len(as_set) > 1 and not link.get("role"):
      link.role = ebgp_role

  """
  build_bgp_sessions: create BGP session data structure

  * BGP route reflectors need IBGP session with all other nodes in the same AS
  * Other nodes need IBGP sessions with all RRs in the same AS
  * EBGP sessions are established whenever two nodes on the same link have different AS
  * Links matching 'advertise_roles' get 'advertise' attribute set
  """
  def build_bgp_sessions(self, node: Box, topology: Box) -> None:
    rrlist = find_bgp_rr(node.bgp.get("as"),topology)
    node.bgp.neighbors = []

    # If we don't have route reflectors, or if the current node is a route
    # reflector, we need BGP sessions to all other nodes in the same AS
    if not(rrlist) or node.bgp.get("rr",None):
      for name,n in topology.nodes.items():
        if "bgp" in n:
          if n.bgp.get("as") == node.bgp.get("as") and n.name != node.name:
            node.bgp.neighbors.append(bgp_neighbor(n,n.loopback,'ibgp',get_neighbor_rr(n)))

    #
    # The node is not a route reflector, and we have a non-empty RR list
    # We need BGP sessions with the route reflectors
    else:
      for n in rrlist:
        if n.name != node.name:
          node.bgp.neighbors.append(bgp_neighbor(n,n.loopback,'ibgp',get_neighbor_rr(n)))

    #
    # EBGP sessions - iterate over all links, find adjacent nodes
    # in different AS numbers, and create BGP neighbors
    for l in node.get("interfaces",[]):
      for ngb_ifdata in l.get("neighbors",[]):
        ngb_name = ngb_ifdata.node
        neighbor = topology.nodes[ngb_name]
        if not "bgp" in neighbor:
          continue
        if neighbor.bgp.get("as") and neighbor.bgp.get("as") != node.bgp.get("as"):
          extra_data = Box({})
          extra_data.ifindex = l.ifindex
          if "unnumbered" in l:
            extra_data.unnumbered = True
            extra_data.local_if = l.ifname

          ebgp_data = bgp_neighbor(neighbor,ngb_ifdata,'ebgp',extra_data)
          if 'vrf' in l:        # VRF neighbor
            if not node.vrfs[l.vrf].bgp.neighbors:
              node.vrfs[l.vrf].bgp.neighbors = []
            node.vrfs[l.vrf].bgp.neighbors.append(ebgp_data)
          else:                 # Global neighbor
            node.bgp.neighbors.append(ebgp_data)

    # Calculate BGP address families
    #
    for af in ['ipv4','ipv6']:
      for n in node.bgp.neighbors:
        if af in n:
          node.bgp[af] = True
          break

  '''
  bgp_set_advertise: set bgp.advertise flag on stub links
  '''
  def bgp_set_advertise(self, node: Box, topology: Box) -> None:
    stub_roles = topology.defaults.bgp.get("advertise_roles",None)
    if 'advertise_roles' in topology.bgp:
      stub_roles = topology.bgp.get("advertise_roles",None)
    if stub_roles:
      for l in node.get("interfaces",[]):
        if "bgp" in l:
          if "advertise" in l.bgp:
            continue
        if l.get("type",None) in stub_roles or l.get("role",None) in stub_roles:
          if not 'bgp' in l:
            l.bgp = {}
          l.bgp.advertise = True

  #
  # Have to set BGP router IDs and cluster IDs before going into node_post_transform
  #
  def module_post_transform(self, topology: Box) -> None:
    for n in topology.nodes.values():
      if 'bgp' in n:
        _routing.router_id(n,'bgp',topology.pools)

    # Build a list of autonomous systems in the lab
    as_list = set( n.bgp['as'] for n in topology.nodes.values() if 'bgp' in n and 'as' in n.bgp )

    # Iterate over all ASNs in the lab, and for route reflectors that don't have rr_cluster_id
    # defined, find an AS-wide cluster ID: the lowest BGP router ID of all route reflectors
    # (members of rr_list) that don't have rr_cluster_id set (because those obviously want to be left alone)
    #
    for asn in as_list:
      rrlist = find_bgp_rr(asn,topology)
      if rrlist:
        rr_cluster_candidates = [n.bgp.router_id for n in rrlist if not 'rr_cluster_id' in n.bgp]
        rr_cluster_id = None
        if rr_cluster_candidates:
          rr_cluster_id = min(rr_cluster_candidates)
        for n in rrlist:
          if not 'rr_cluster_id' in n.bgp:
            n.bgp.rr_cluster_id = rr_cluster_id or n.bgp.router_id
          elif n.bgp.rr_cluster_id:
            try:
              n.bgp.rr_cluster_id = str(netaddr.IPAddress(n.bgp.rr_cluster_id).ipv4())
            except Exception as ex:
              common.error(
                f'BGP cluster ID {n.bgp.rr_cluster_id} configured on node {n.name} cannot be converted into an IPv4 address',
                common.IncorrectValue,
                'bgp')

  #
  # Execute the rest of node post-transform code, including setting up the BGP session table
  #
  def node_post_transform(self, node: Box, topology: Box) -> None:
    if not "bgp" in node:   # pragma: no cover (this should have been caught in check_bgp_parameters)
      common.fatal(f"Internal error: node {node.name} has BGP module enabled but no BGP parameters","bgp")
      return
    check_bgp_parameters(node)
    self.build_bgp_sessions(node,topology)
    self.bgp_set_advertise(node,topology)



#
# Dump transformed topology in JSON format
#
# This is just a placeholder module (because we have to load something),
# the real work is done in the YAML module which tests for class name and
# prints out YAML or JSON string
#
from .yaml import YAML

class JSON(YAML):
  pass



#
# netlab up command
#
# * Transform lab topology and create provider and automation files, 
#   or read transformed lab topology from snapshot file
# * Start the lab, including provider-specific pre- and post-start hooks
#
import typing
import argparse
import os
import glob
import subprocess
import shutil

from box import Box
from pathlib import Path

from .. import common
from . import create
from . import external_commands
from . import common_parse_args, load_snapshot_or_topology
from .. import providers
from .. import read_topology

#
# Extra arguments for 'netlab up' command
#
def up_parse_args(standalone: bool) -> argparse.ArgumentParser:
  parse_parents = [ common_parse_args() ] if standalone else []
  parser = argparse.ArgumentParser(
    description='netlab up extra arguments',
    parents = parse_parents,
    add_help=standalone)
  parser.add_argument(
    '--no-config',
    dest='no_config',
    action='store_true',
    help='Do not configure lab devices')
  parser.add_argument(
    '--fast-config',
    dest='fast_config',
    action='store_true',
    help='Use fast device configuration (Ansible strategy = free)')
  parser.add_argument(
    '--snapshot',
    dest='snapshot',
    action='store',
    nargs='?',
    const='netlab.snapshot.yml',
    help='Use netlab snapshot file created by a previous lab run')
  return parser

def run(cli_args: typing.List[str]) -> None:
  up_args_parser = up_parse_args(False)                       # Try to parse the up-specific arguments
  (args,rest) = up_args_parser.parse_known_args(cli_args)

  if args.snapshot:                                           # If we're using the snapshot file...
    up_args_parser = up_parse_args(True)                      # ... create a stricter parser with help option
    args = up_args_parser.parse_args(cli_args)                # ... and reparse
    common.set_logging_flags(args)                            # ... use these arguments to set logging flags and read the snapshot

    topology = read_topology.read_yaml(filename=args.snapshot)
    if topology is None:
      common.fatal('Cannot read snapshot file {args.snapshot}, aborting...')
      return

    print(f"Using transformed lab topology from snapshot file {args.snapshot}")
  else:                                                       # No snapshot file, use 'netlab create' parser
    topology = create.run(cli_args,'up','Create configuration files, start a virtual lab, and configure it',up_args_parser)

  settings = topology.defaults
  if common.QUIET:
    os.environ["ANSIBLE_STDOUT_CALLBACK"] = "selective"

  external_commands.run_probes(settings,topology.provider,2)

  provider = providers._Provider.load(topology.provider,topology.defaults.providers[topology.provider])
  provider.call('pre_output_transform',topology)

  if hasattr(provider,'pre_start_lab') and callable(provider.pre_start_lab):
    provider.pre_start_lab(topology)

  external_commands.start_lab(settings,topology.provider,3,"netlab up")
  
  if hasattr(provider,'post_start_lab') and callable(provider.post_start_lab):
    provider.post_start_lab(topology)

  if not args.no_config:
    external_commands.deploy_configs(4,"netlab up",args.fast_config)
  else:
    print("\nInitial configuration skipped, run 'netlab initial' to configure the devices")



#
# Dynamic output framework
# 
# Individual output routines are defined in modules within this directory inheriting
# TopologyOutput class and replacing or augmenting its methods (most commonly, write)
#

import platform
import subprocess
import os
import typing
import re

# Related modules
from box import Box

from .. import common
from ..callback import Callback

class _TopologyOutput(Callback):
  def __init__(self, output: str, data: Box) -> None:
    self.settings = data

    output_parts = output.split('=')

    output_list = output_parts[0].split(':')
    self.output = output_list[0]
    self.format = output_list[1:]

    if len(output_parts) > 1:
      self.filenames = output_parts[1].split(',')

  @classmethod
  def load(self, output: str, data: Box) -> typing.Optional['_TopologyOutput']:
    module_name = __name__+"."+re.split(':|=',output)[0]
    obj = self.find_class(module_name)
    if obj:
      return obj(output,data)
    else:
      return None

  def get_template_path(self) -> str:
    fmtname = self.format[0] if self.format else "default"
    return 'templates/outputs/' + self.output + "/" + fmtname + ".j2"

  def get_output_name(self, fname: str, topology: Box) -> typing.Optional[str]:
    if fname:
      return fname

    if self.settings:
      return self.settings.filename

    return None

  def write(self, topology: Box) -> None:
    common.fatal('someone called the "write" method of TopologyOutput abstract class')



#
# SRv6 transformation module
#
from box import Box

from . import _Module
from .. import addressing, common
import ipaddr # type: ignore

class SRV6(_Module):

  def node_post_transform(self, node: Box, topology: Box) -> None:

      # Could model this as another addressing pool too
      locator = ipaddr.IPv6Network( f'{topology.defaults.srv6.locator}:{node.id:x}::/64' )

      if 'ipv6' not in node.loopback:
          common.error( "SRv6 requires an ipv6 loopback address",
                        common.MissingValue, 'srv6' )
      elif locator.overlaps( ipaddr.IPNetwork(node.loopback.ipv6) ):
          common.error( f"Node ipv6 loopback address {node.loopback.ipv6} overlaps with locator {locator}",
                        common.IncorrectValue, 'srv6' )

      node.srv6.locator = str( locator )

      # TODO process per-interface srv6 parameters?
      # for l in node.get('interfaces',[]):



#
# VRF module
#
import typing
from box import Box

from . import _Module,_routing
from .. import common
from .. import data
from ..data import get_from_box
from ..augment import devices

vrf_id_set: set
vrf_last_id: int

#
# build_vrf_id_set: given an object (topology or node), create a set of RDs
# that appear in that object.
#
def build_vrf_id_set(obj: Box, attr: str) -> set:
  if 'vrfs' in obj:
    return { v[attr] for v in obj.vrfs.values() if isinstance(v,dict) and attr in v and v[attr] is not None }
  return set()

def populate_vrf_id_set(topology: Box) -> None:
  global vrf_id_set, vrf_last_id

  vrf_id_set = build_vrf_id_set(topology,'rd')
  vrf_last_id = 1

  for n in topology.nodes.values():
    if 'vrfs' in n:
      vrf_id_set = vrf_id_set.union(build_vrf_id_set(n,'rd'))

#
# Get a usable AS number. Try bgp.as then vrf.as from node and global settings
#
def get_rd_as_number(obj: Box, topology: Box) -> typing.Optional[typing.Any]:
  return \
    get_from_box(obj,'bgp.as') or \
    get_from_box(obj,'vrf.as') or \
    get_from_box(topology,'bgp.as') or \
    get_from_box(topology,'vrf.as')

#
# Parse rd/rt value -- check whether the RD/RT value is in N:N format
#

def parse_rdrt_value(value: str) -> typing.Optional[typing.List[int]]:
  try:
    (asn,vid) = str(value).split(':')
    return [int(asn),int(vid)]
  except Exception as ex:
    return None

def get_next_vrf_id(asn: str) -> str:
  global vrf_id_set, vrf_last_id

  while f'{asn}:{vrf_last_id}' in vrf_id_set:
    vrf_last_id = vrf_last_id + 1

  rd = f'{asn}:{vrf_last_id}'
  vrf_id_set.add(rd)
  return rd

#
# Normalize VRF IDs -- give a set of VRFs, change integer values of RDs into N:N strings
#

def normalize_vrf_dict(obj: Box, topology: Box) -> None:
  if not 'vrfs' in obj:
    return

  asn = None
  obj_name = 'global VRFs' if obj is topology else obj.name

  if not isinstance(obj.vrfs,dict):
    common.error(f'VRF definition in {obj_name} is not a dictionary',common.IncorrectValue,'vrf')
    return

  for vname in list(obj.vrfs.keys()):
    if obj.vrfs[vname] is None:
      obj.vrfs[vname] = {}
    if not isinstance(obj.vrfs[vname],dict):
      common.error(f'VRF definition for {vname} in {obj_name} should be empty or a dictionary',
        common.IncorrectValue,
        'vrf')
      return

    vdata = obj.vrfs[vname]
    if 'rd' in vdata:
      if vdata.rd is None:      # RD set to None can be used to auto-generate RD while preventing RD inheritance
        continue                # ... skip the rest of the checks
      if isinstance(vdata.rd,int):
        asn = asn or get_rd_as_number(obj,topology)
        if not asn:
          common.error(f'VRF {vname} in {obj_name} uses integer RD value without a usable vrf.as or bgp.as AS number',
            common.MissingValue,
            'vrf')
          return
        vdata.rd = f'{asn}:{vdata.rd}'
      elif isinstance(vdata.rd,str):
        if parse_rdrt_value(vdata.rd) is None:
          common.error(f'RD value in VRF {vname} in {obj_name} is not in N:N format',
            common.IncorrectValue,
            'vrf')
      else:
        common.error(f'RD value in VRF {vname} in {obj_name} must be a string or an integer',
          common.IncorrectValue,
          'vrf')

def normalize_vrf_ids(topology: Box) -> None:
  normalize_vrf_dict(topology,topology)

  for n in topology.nodes.values():
    normalize_vrf_dict(n,topology)

#
# Get VRF RD value needed for import/export values. 
#
# WARNING: Global value takes precedence over node value because you might want to change per-node RD
# values for weird topologies like hub-and-spoke
#
def get_vrf_id(vname: str, obj: Box, topology: Box) -> typing.Optional[str]:
  obj_name = 'global VRFs' if obj is topology else obj.name
  vdata = get_from_box(topology,['vrfs',vname]) or get_from_box(obj,['vrfs',vname]) or None

  if vdata is None:
    common.error(
      f'Cannot get VRF ID for unknown VRF {vname} needed in {obj_name}',
      common.MissingValue,
      'vrf')
    return None

  if not isinstance(vdata,Box):
    common.fatal(f'Internal error: got a VRF definition that is not a dictionary')
    return None

  if not 'rd' in vdata:
    common.fatal(f'Internal error: VRF {vname} in {obj_name} should have a RD value by now')
    return None

  return vdata.rd

#
# Set RD values for all VRFs that have no RD attribute or RD value set to None (= auto-generate)
#
def set_vrf_ids(obj: Box, topology: Box) -> None:
  if not 'vrfs' in obj:
    return

  asn = None
  obj_name = 'global VRFs' if obj is topology else obj.name

  for vname,vdata in obj.vrfs.items():
    if vdata.get('rd',None) is None:
      asn = asn or get_rd_as_number(obj,topology)
      if not asn:
        common.error('Need a usable vrf.as or bgp.as to create auto-generated VRF RD for {vname} in {obj_name}',
          common.MissingValue,
          'vrf')
        return
      vdata.rd = get_next_vrf_id(asn)

#
# Set import/export route targets
#
def set_import_export_rt(obj : Box, topology: Box) -> None:
  if not 'vrfs' in obj:
    return None

  obj_name = 'global VRFs' if obj is topology else obj.name
  obj_id   = 'vrfs' if obj is topology else f'nodes.{obj.name}.vrfs'
  asn      = None

  for vname,vdata in obj.vrfs.items():
    for rtname in ['import','export']:
      if not rtname in vdata:
        vdata[rtname] = [ vdata.rd ]
        continue

      data.must_be_list(vdata,rtname,f'{obj_id}.{vname}')

      rtlist = []     # The final parsed and looked-up list of RT values
      for rtvalue in vdata[rtname]:
        if isinstance(rtvalue,int):         # RT can be specified as an integer, in which case ASN is prepended to it
          asn = asn or get_rd_as_number(obj,topology)
          if not asn:
            common.error('VRF {vname} in {obj_id} uses integer {rtname} value without a usable vrf.as or bgp.as AS number',
              common.MissingValue,
              'vrf')
            continue
          rtvalue = f'{asn}:{rtvalue}'
        elif not isinstance(rtvalue,str):   # If RT is not an integer, it really should be a string
          common.error('{rtname} value {rtvalue} in VRF {vname} in {obj_id} should be a string or an integer',
            common.IncorrectValue,
            'vrf')
          continue
        else:
          if ':' in rtvalue:                # If there's a colon in RT value, then we're assuming N:N format
            if parse_rdrt_value(rtvalue) is None:
              common.error('{rtname} value {rtvalue} in VRF {vname} in {obj_id} is not in valid N:N format',
                common.IncorrectValue,
                'vrf')
              continue
          else:                             # Otherwise the RT value should refer to another VRF name
            rtvalue = get_vrf_id(rtvalue,obj,topology)
            if rtvalue is None:
              continue            # Error message generated in get_vrf_id

        rtlist.append(rtvalue)

      vdata[rtname] = rtlist

#
# VRF route leaking is usually implemented through BGP VPNv4 address families
# Check whether we have BGP AS configured on all nodes that use VRFs with route leaking
# (identified as import or export RT not equal to [ RD ])
#

def validate_vrf_route_leaking(node : Box) -> None:
  for vname,vdata in node.vrfs.items():
    simple_rt = [ vdata.rd ]
    leaked_routes = vdata['import'] and vdata['import'] != simple_rt
    leaked_routes = leaked_routes or (vdata['export'] and vdata['export'] != simple_rt)
    if leaked_routes:
      if not get_from_box(node,'bgp.as'):
        if get_from_box(node,'vrf.as'):
          node.bgp['as'] = node.vrf['as']
        else:
          common.error(
            f"VRF {vname} on {node.name} uses inter-VRF route leaking, but there's no BGP AS configured on the node",
            common.MissingValue,
            'vrf')

class VRF(_Module):

  def module_pre_default(self, topology: Box) -> None:
    for attr_set in ['global','node']:
      if not 'vrfs' in topology.defaults.attributes[attr_set]:
        topology.defaults.attributes[attr_set].append('vrfs')

  def module_pre_transform(self, topology: Box) -> None:
    if not 'vrfs' in topology:
      return

    normalize_vrf_ids(topology)
    populate_vrf_id_set(topology)
    set_vrf_ids(topology,topology)
    set_import_export_rt(topology,topology)

  def node_pre_transform(self, node: Box, topology: Box) -> None:
    if not 'vrfs' in node:
      return

    for vname in node.vrfs.keys():
      if 'vrfs' in topology and vname in topology.vrfs:
        node.vrfs[vname] = topology.vrfs[vname] + node.vrfs[vname]

    set_vrf_ids(node,topology)
    set_import_export_rt(node,topology)

  def link_pre_transform(self, link: Box, topology: Box) -> None:
    pass

  def node_post_transform(self, node: Box, topology: Box) -> None:
    vrf_count = 0
    for ifdata in node.interfaces:
      if 'vrf' in ifdata:
        vrf_count = vrf_count + 1
        if 'vrfs' in topology and ifdata.vrf in topology.vrfs:
          node.vrfs[ifdata.vrf] = topology.vrfs[ifdata.vrf] + node.vrfs[ifdata.vrf]

        if not node.vrfs[ifdata.vrf]:
          common.error(
            f'VRF {ifdata.vrf} used on an interface in {node.name} is not defined in the node or globally',
            common.MissingValue,
            'vrf')
          continue
        if not node.vrfs[ifdata.vrf].rd:
          common.error(
            f'VRF {ifdata.vrf} used on an interface in {node.name} does not have a usable RD',
            common.MissingValue,
            'vrf')
          continue

        for af in ['v4','v6']:
          if f'ip{af}' in ifdata:
            node.af[f'vpn{af}'] = True
            node.vrfs[ifdata.vrf].af[f'ip{af}'] = True

    if not vrf_count:                # Remove VRF module from the node if the node has no VRFs
      node.module = [ m for m in node.module if m != 'vrf' ]
    else:
      node.vrfs = node.vrfs or {}     # ... otherwise make sure the 'vrfs' dictionary is not empty
      vrfidx = 100
      for v in node.vrfs.values():    # We need unique VRF index to create OSPF processes
        v.vrfidx = vrfidx
        vrfidx = vrfidx + 1

      validate_vrf_route_leaking(node)

    # Finally, set BGP router ID if we set BGP AS number
    #
    if get_from_box(node,'bgp.as') and not get_from_box(node,'bgp.router_id'):
      _routing.router_id(node,'bgp',topology.pools)



'''
Set up network addressing
=========================

The network topology needs three types of addressing pools:

* Loopback addresses
* P2P addresses
* LAN and stub link addresses

Each pool could have IPv4 and IPv6 component. Alternatively, a pool
could be unnumbered (using IPv6 LLA and IPv4 address from loopback interface).

Notes:
* _lan_ pool is used for stub and P2P prefixes if the corresponding
  pools are not defined

The addressing pools are set up in _addressing_ part of topology
defaults (local or global), for example:

addressing:
  lan:
    ipv4: 172.16.0.0/16
    ipv6: 2008:db8:0100::/48
    prefix: 24
  p2p:
    ipv4: 10.1.0.0/16
    prefix: 30
  loopback:
    ipv4: 10.0.0.0/24

IPv4 component of an addressing pool is defined with two parameters:

* _ipv4_ prefix - IPv4 addressing space used by the pool
* _prefix_ length - prefix length of individual subnets in the pool

IPv6 component of an addressing pool is defined with an _ipv6_ prefix,
prefix length of individual subnets is assumed to be /64.

Legacy setup migration
----------------------
While building the addressing pools the **setup** routine uses legacy
parameters when the corresponding addressing parameteres are missing:

* _loopback_ default is used to create _loopback_ pool
* _lan_ and _lan_subnet_ defaults are used to create _LAN_ pool
* _p2p_ and _p2p_subnet_ defaults are used to create _p2p_ pool
'''

import sys
import typing

import netaddr
from box import Box

# Related modules
from . import common
from . import data

def normalize_prefix(pfx: typing.Union[str,Box]) -> Box:
  if not pfx:
    return Box({},default_box=True,box_dots=True)
  if not isinstance(pfx,dict):
    return Box({ 'ipv4': str(pfx)},default_box=True,box_dots=True)
  for af in 'ipv4','ipv6':
    if af in pfx:
      if not pfx[af] or 'unnumbered' in pfx:
        del pfx[af]

  return pfx

def rebuild_prefix(pfx: typing.Union[dict,Box]) -> dict:
  out_pfx = {}
  for af in ('ipv4','ipv6'):
    if af in pfx:
      out_pfx[af] = str(pfx[af])
  return out_pfx

def setup_pools(addr_pools: typing.Optional[Box] = None, defaults: typing.Optional[Box] = None) -> Box:
  addrs = addr_pools or Box({},default_box=True)
  defaults = defaults or Box({},default_box=True)
  legacy = Box({},default_box=True)

  legacy.lan = { 'ipv4': defaults.get('lan','10.0.0.0/16'), 'prefix': defaults.get('lan_subnet',24) }
  if not 'lan' in defaults and not 'lan' in addrs:
    legacy['lan']['start'] = 1
  legacy.loopback = { 'ipv4': (defaults.get('loopback','10.0.0.%d') % 0) + '/24', 'prefix': 32 }
  legacy.p2p = { 'ipv4': defaults.get('p2p','10.2.0.0/16'), 'prefix': defaults.get('p2p_subnet',30) }

  addrs = legacy + addrs

  # Replace string pool definitions with data structures
  for pool in list(addrs):
    clean_pfx = normalize_prefix(addrs[pool])
    if 'ipv4' in clean_pfx and not 'prefix' in clean_pfx:
      clean_pfx.prefix = 32 if 'loopback' in pool else 24
    addrs[pool] = clean_pfx

  return addrs

def validate_pools(addrs: typing.Optional[Box] = None) -> None:
  if not addrs:       # pragma: no cover (pretty hard not to have address pools)
    addrs = Box({})
  for k in ('lan','loopback'):
    if not k in addrs:          # pragma: no cover (lan and loopback pools are always created in setup_pools)
      common.error(
        "'%s' addressing pool is missing" % k,
        category=common.MissingValue,
        module='addressing')

  if isinstance(addrs.mgmt,dict):
    if not 'prefix' in addrs.mgmt:
      addrs.mgmt['prefix'] = 24

  for pool,pfx in addrs.items():
    if 'unnumbered' in pfx:
      if 'ipv4' in pfx or 'ipv6' in pfx:    # pragma: no cover -- ipv4/ipv6 prefixes have already been removed from unnumbered pools
        common.error(
          f'Pool {pool} is an unnumbered pool and cannot have IPv4 or IPv6 prefixes {pfx}',
          category=common.IncorrectValue,
          module='addressing')
        continue
    for k in ('ipv4','ipv6'):
      if k in pfx:
        if not isinstance(pfx[k],bool):
          try:
            network = netaddr.IPNetwork(pfx[k])
            addrs[pool][k+'_pfx'] = network
          except:
            common.error(
              "%s prefix %s in addressing pool '%s' is invalid (%s)" % (k,pfx[k],pool,sys.exc_info()[1]),
              category=common.IncorrectValue,
              module='addressing')
            continue

    if 'mac' in pfx:
      try:
        addrs[pool].mac_eui = netaddr.EUI(pfx.mac)
      except:
        common.error(
          "MAC prefix %s in addressing pool '%s' is invalid (%s)" % (pfx.mac,pool,sys.exc_info()[1]),
          category=common.IncorrectValue,
          module='addressing')
        continue

    if 'ipv4' in pfx and 'ipv4_pfx' in pfx and pool != 'mgmt':
      if not 'prefix' in pfx:   # pragma: no cover -- default prefix was already set to /24
        common.error(
          "IPv4 prefix length is missing in '%s' addressing pool" % pool,
          category=common.MissingValue,
          module='addressing')
      else:
        if not isinstance(pfx['prefix'],int):
          common.error(
            "IPv4 prefix length in '%s' addressing pool is not an integer" % pool,
            category=common.IncorrectValue,
            module='addressing')
        else:
          if pfx.prefix > 32 or pfx.prefix < 1:
            common.error(
              "IPv4 subnet prefix length in '%s' addressing pool is not between 1 and 32" % pool,
              category=common.IncorrectValue,
              module='addressing')
          if pfx.prefix < pfx.ipv4_pfx.prefixlen:
            common.error(
              "IPv4 subnet prefix length in '%s' addressing pool is not longer than pool prefix" % pool,
              category=common.IncorrectValue,
              module='addressing')

    if 'ipv6' in pfx and 'ipv6_pfx' in pfx:
      if pfx.ipv6_pfx.prefixlen > 56:
        common.error(
          "Error in '%s' addressing pool: IPv6 pool prefix cannot be longer than /56" % pool,
          category=common.IncorrectValue,
          module='addressing')

  if not 'ipv4' in addrs.loopback and not 'ipv6' in addrs.loopback:
    common.error(
      "Loopback addressing pool has no IPv4 or IPv6 address prefix",
      category=common.MissingValue,
      module='addressing')

def create_pool_generators(addrs: typing.Optional[Box] = None) -> typing.Dict:
  if not addrs:       # pragma: no cover (pretty hard not to have address pools)
    addrs = Box({})
  gen: typing.Dict = {}
  for pool,pfx in addrs.items():
    gen[pool] = {}
    for key,data in pfx.items():
      if "_pfx" in key:
        af   = key.replace('_pfx','')
        plen = pfx['prefix'] if af == 'ipv4' else 64
        gen[pool][af] = data.subnet(plen)
        if (af == 'ipv4' and plen == 32) or (pool == 'loopback'):
          next(gen[pool][af])
      elif isinstance(data,bool):
        gen[pool][key] = data
  return gen

def get_pool(pools: Box, pool_list: typing.List[str]) -> typing.Optional[str]:
  for p in pool_list:
    if p in pools:
      return p

  common.error(
    f'Cannot get addressing for any of these pools: {pool_list}',
    category=common.MissingValue,
    module='addressing')                       # pragma: no cover (impossible to get here due to built-in default pools)
  return None                                  # pragma: no cover

def get_nth_subnet(n: int, subnet: netaddr.IPNetwork.subnet, cache_list: list) -> netaddr.IPNetwork:
  while len(cache_list) < n:
    cache_list.append(next(subnet))
  return cache_list[n-1]

def get_pool_prefix(pools: typing.Dict, p: str, n: typing.Optional[int] = None) -> typing.Dict:
  prefixes: typing.Dict = {}
  if pools[p].get('unnumbered'):
    return { 'unnumbered': True }
  for af in list(pools[p]):
    if not 'cache' in af:
      if isinstance(pools[p][af],bool):
        prefixes[af] = pools[p][af]
      elif n:
        subnet_cache = 'cache_%s' % af
        if not subnet_cache in pools[p]:
          pools[p][subnet_cache] = []
        try:
          prefixes[af] = get_nth_subnet(n,pools[p][af],pools[p][subnet_cache])
        except StopIteration:
          common.error(
            f'Cannot allocate {n}-th {af} element from {p} pool',
            common.IncorrectValue,
            'addressing')
      else:
        try:
          prefixes[af] = next(pools[p][af])
        except StopIteration:
          common.error(
            f'Ran out of {af} prefixes in {p} pool' +
            (' (use --debug flag to get more details)' if not common.DEBUG else ''),
            common.MissingValue,
            'addressing')

  return prefixes

def get(pools: Box, pool_list: typing.Optional[typing.List[str]] = None, n: typing.Optional[int] = None) -> typing.Dict:
  if not pool_list:
    pool_list = ['lan']                   # pragma: no cover
  p = get_pool(pools,pool_list)
  if p:
    return get_pool_prefix(pools,p,n)
  else:
    return {}                             # pragma: no cover -- can't figure out how to get here

def setup(topo: Box, defaults: Box) -> None:
  data.null_to_string(topo.addressing)
  addrs = setup_pools(defaults.addressing + topo.addressing,defaults)

  common.print_verbose("Addressing\n=================")
  common.print_verbose(addrs.to_yaml())

  validate_pools(addrs)
  common.exit_on_error()

  topo.pools = create_pool_generators(addrs)
  topo.addressing = addrs
  common.print_verbose("Pools\n=================")
  common.print_verbose(str(topo.pools))

  common.exit_on_error()

def parse_prefix(prefix: typing.Union[str,dict]) -> typing.Dict:
  if common.DEBUG:                     # pragma: no cover (debugging printout)
    print(f"parse prefix: {prefix}")
  if not prefix:
    return {}
  supported_af = ['ip','ipv4','ipv6']
  prefix_list: typing.Dict = {}
  if isinstance(prefix,dict):
    for af,pfx in prefix.items():
      if not af in supported_af:
        common.error( \
          'Unknown address family %s in prefix %s' % (af,prefix), \
          category=common.IncorrectValue,module='addressing')
      else:
        if isinstance(pfx,bool):
          prefix_list[af] = pfx
        else:
          try:
            prefix_list[af] = netaddr.IPNetwork(pfx)
          except Exception as ex:
            common.error(f'Cannot parse {af} prefix: {prefix}\n... {ex}',common.IncorrectValue,'addressing')
            return {}
          if str(prefix_list[af]) != str(prefix_list[af].cidr):
            common.error(f'{af} prefix contains host bits: {prefix}',common.IncorrectValue,'addressing')
    return prefix_list
  else:
    return { 'ipv4' : netaddr.IPNetwork(prefix) }

def get_addr_mask(pfx: netaddr.IPNetwork, host: int) -> str:
  host_ip = netaddr.IPNetwork(pfx[host])
  host_ip.prefixlen = pfx.prefixlen
  return str(host_ip)



#
# Create detailed link data structures including automatic interface numbering
# from high-level topology
#
import typing

import netaddr
from box import Box

# Related modules
from .. import common
from .. import data
from .. import addressing
from . import devices

IFATTR: str = 'interfaces'          # Just in case we decide to call it something else not 'interfaces'

def adjust_interface_list(iflist: list, link: typing.Any, nodes: Box) -> list:
  link_intf = []
  for n in iflist:                      # Sanity check of interface data
    if isinstance(n,str):               # Another shortcut: node name as string
      n = Box({ 'node': n },default_box=True,box_dots=True)
    if not isinstance(n,dict):          # Still facing non-dict data type?
      common.error(                     # ... report an error
        f'Interface description {n} on link {link} must be a dictionary',
        common.IncorrectValue,
        'links')
    elif not 'node' in n:               # Do we have node name in interface data?
      common.error(                     # ... no? Too bad, throw an error
        f'Interface data {n} on link {link} is missing a "node" attribute',
        common.MissingValue,
        'links')
    elif not n['node'] in nodes:        # Is the node name valid?
      common.error(                     # ... it's not, get lost
        f'Interface data {n} on link {link} refers to an unknown node {n["node"]}',
        common.IncorrectValue,
        'links')
    else:
      link_intf.append(n)               # Interface data is OK, append it to interface list

  return link_intf

def adjust_link_list(links: list, nodes: Box) -> list:
  global IFATTR
  link_list: list = []

  if not(links):
    return link_list

  link_cnt = 0
  for l in links:
    if isinstance(l,dict) and IFATTR in l:               # a dictionary with 'interfaces' element
      l = Box(l,default_box=True,box_dots=True)
      data.must_be_list(l,IFATTR,f'link[{link_cnt}]')    # ... check it's a list and move on
      l[IFATTR] = adjust_interface_list(l[IFATTR],l,nodes)
      link_list.append(l)
    elif isinstance(l,dict):                             # a dictionary without 'interfaces' element
      link_data = {}                                     # ... split it into link attributes
      link_intf = []                                     # ... and a list of nodes
      for (k,v) in l.items():
        if k in nodes:                                   # Node name -> interface list
          if not v:
            v = Box({},default_box=True,box_dots=True)
          if not isinstance(v,dict):                     # Interface data must be a dictionary
            common.error(
              f'Interface data {v} for node {k} on link {l} must be a dictionary',
              common.IncorrectValue,
              'links')
            continue
          v['node'] = k                                  # ... add 'node' to the interface so we know what node it belongs to
          link_intf.append(v)
        else:
          link_data[k] = v                  # ... otherwise copy key/value pair to link data
      link_data[IFATTR] = link_intf         # Add revised interface data to link data
      link_list.append(link_data)           # ... and move on
    elif isinstance(l,list):
      link_list.append(Box({ IFATTR: adjust_interface_list(l,l,nodes) },default_box=True,box_dots=True))
    else:                                   # Assuming the link value is a string, split
      link_intf = []
      for n in l.split('-'):                # ... split it into a list of nodes
        if n in nodes:                      # If the node name is valid
          link_intf.append({ 'node': n })   # ... append it to the list of interfaces
        else:
          common.error(
            f'Link string {l} refers to an unknown node {n}',
            common.IncorrectValue,
            'links')
      link_list.append({ IFATTR: link_intf })
    link_cnt = link_cnt + 1

  if common.DEBUG:
    print("Adjusted link list")
    print("=" * 60)
    print(common.print_yaml(link_list))

  return link_list

def get_link_full_attributes(defaults: Box) -> set:
  attributes = defaults.get('attributes',{})
  user = attributes.get('link',[])
  internal = attributes.get('link_internal',[])

  set_attributes = set(user).union(set(internal))
  if 'module' in defaults:
    set_attributes = set_attributes.union(set(defaults.module))

  if 'link_attr' in defaults:
    set_attributes = set_attributes.union(set(defaults.link_attr))

  return set_attributes

def get_link_base_attributes(defaults: Box) -> set:
  attributes = defaults.get('attributes',{})
  no_propagate = attributes.get('link_no_propagate')
  return get_link_full_attributes(defaults) - set(no_propagate)

def add_node_interface(node: Box, ifdata: Box, defaults: Box) -> Box:
  ifindex_offset = devices.get_device_attribute(node,'ifindex_offset',defaults)
  if ifindex_offset is None:
    ifindex_offset = 1

  ifindex = len(node.interfaces) + ifindex_offset

  ifname_format = devices.get_device_attribute(node,'interface_name',defaults)

  ifdata.ifindex = ifindex
  if ifname_format:
    ifdata.ifname = ifname_format % ifindex

  pdata = devices.get_provider_data(node,defaults).get('interface',{})
  pdata = Box(pdata,box_dots=True,default_box=True)                     # Create a copy of the provider interface data
  if 'name' in pdata:
    pdata.name = pdata.name % ifindex

  if pdata:
    provider = devices.get_provider(node,defaults)
    ifdata[provider] = pdata

  for af in ('ipv4','ipv6'):
    if af in ifdata and not ifdata[af]:
      del ifdata[af]

  if 'mtu' in node:                             # Is node-level MTU defined (node setting, lab default or device default)
    sys_mtu = devices.get_device_features(node,defaults).initial.get('system_mtu',False)
    if 'mtu' in ifdata:                         # Is MTU defined on the interface?
      if sys_mtu and node.mtu == ifdata.mtu:    # .. is it equal to node MTU?
        ifdata.pop('mtu',None)                  # .... remove interface MTU on devices that support system MTU
    else:                                       # Node MTU is defined, interface MTU is not
      if not sys_mtu:                           # .. does the device support system MTU?
        ifdata.mtu = node.mtu                   # .... no, copy node MTU to interface MTU

  node.interfaces.append(ifdata)

  # Box modifies the dict in place, return a reference to be updated
  # return len(node.links)
  return node.interfaces[-1]

# Add common interface data to node ifaddr structure
#
def interface_data(link: Box, link_attr: set, ifdata: Box) -> Box:
  for k in link_attr:
    if k in link:
      if not k in ifdata:
        ifdata[k] = link[k]
      elif isinstance(link[k],dict) and isinstance(ifdata[k],dict):
        ifdata[k] = link[k] + ifdata[k]
  return ifdata

#
# Add module-specific data to node ifaddr structure
#
# Iterate over modules, for every matching key in link definition
# copy the value into node ifaddr
#
def ifaddr_add_module(ifaddr: Box, link: Box, module: Box) -> None:
  if module:
    for m in module:
      if m in link:
        ifaddr[m] = link[m]

#
# get_node_link_address: calculate on-link address for the specified node
#
# node: node data collected so far
# ifaddr: interface data collected so far
# node_link_data: existing node-on-link data
# prefix: link prefix (ipv4, ipv6, unnumbered)
# node_id: desired address within the subnet
#

def get_node_link_address(node: Box, ifdata: Box, node_link_data: dict, prefix: dict, node_id: int) -> typing.Optional[str]:
  if common.DEBUG:     # pragma: no cover (debugging)
    print(f"get_node_link_address for {node.name}:\n"+
          f".. ifdata: {ifdata}\n"+
          f".. node_link_data: {node_link_data}\n"+
          f".. prefix: {prefix}\n"+
          f".. node_id: {node_id}")
  if 'unnumbered' in prefix:          # Special case: old-style unnumbered link
    if common.DEBUG:     # pragma: no cover (debugging)
      print(f"... node loopback: {node.loopback}")
    for af in ('ipv4','ipv6'):        # Set AF to True for all address families
      if af in node_link_data:
        return(f'{af} address ignored for node {node.name} on a fully-unnumbered link')
      if af in node.loopback:         # ... present on node loopback interface
        ifdata[af] = True
        node_link_data[af] = True
    return None

  for af in ('ipv4','ipv6'):
    node_addr = None
    if af in node_link_data:                  # static IP address or host index
      if isinstance(node_link_data[af],bool): # unnumbered node, leave it alone
        continue
      if isinstance(node_link_data[af],int):  # host portion of IP address specified as an integer
        if af in prefix:
          if isinstance(prefix[af],bool):
            return(f'Node {node.name} is using host index for {af} on an unnumbered link')
          try:
            node_addr = netaddr.IPNetwork(prefix[af][node_link_data[af]])
            node_addr.prefixlen = prefix[af].prefixlen
          except Exception as ex:
            return(
              f'Cannot assign host index {node_link_data[af]} in {af} from prefix {prefix[af]} to node {node.name}\n'+
              f'... {ex}')
        else:
          return(f'Node {node.name} is using host index {node_link_data[af]} for {af} on a link that does not have {af} prefix')
      else:                                  # static IP address
        try:
          node_addr = netaddr.IPNetwork(node_link_data[af])
        except:
          return(f'Invalid {af} link address {node_link_data[af]} for node {node.name}')
        if '/' not in node_link_data[af] and af in prefix:
          node_addr.prefixlen = prefix[af].prefixlen
        if str(node_addr) == str(node_addr.cidr):        # Check whether the node address includes a host portion
          lb = not(':' in str(node_addr)) \
                 and node_addr.prefixlen == 32           # Exception#1: IPv4/32
          lb = lb or node_addr.prefixlen == 128          # Exception#2: IPv6/128
          if not lb:
            return(f'Static node address {node_link_data[af]} for node {node.name} does not include a host portion')
    elif af in prefix:
      if isinstance(prefix[af],bool):        # New-style unnumbered link
        if prefix[af]:                       # Copy only True value into interface data
          ifdata[af] = prefix[af]            # ... to ensure AF presence in ifdata indicates protocol-on-interface
          node_link_data[af] = prefix[af]
      else:
        try:
          index = node_id-1 if af == 'ipv4' and prefix[af].prefixlen==31 else node_id
          node_addr = netaddr.IPNetwork(prefix[af][index])
        except Exception as ex:
          return(
            f'Cannot assign {af} address from prefix {prefix[af]} to node {node.name} with ID {node.id}\n'+
            f'... {ex}')
        node_addr.prefixlen = prefix[af].prefixlen

    if node_addr:
      node_link_data[af] = str(node_addr)
      ifdata[af] = node_link_data[af]

  if common.DEBUG:     # pragma: no cover (debugging)
    print(f"get_node_link_address for {node.name} completed:\n"+
          f".. ifdata: {ifdata}\n"+
          f".. node_link_data: {node_link_data}")
    print
  return None

def augment_link_prefix(link: Box,pools: typing.List[str],addr_pools: Box) -> dict:
  if 'role' in link:
    pools = [ link.get('role') ] + pools
  if 'prefix' in link:
    pfx_list = addressing.parse_prefix(link.prefix)
    if isinstance(link.prefix,str):
      link.prefix = addressing.rebuild_prefix(pfx_list)
  elif 'unnumbered' in link:
    pfx_list = Box({ 'unnumbered': True })
  else:
    pfx_list = addressing.get(addr_pools,pools)
    link.prefix = {
        af: pfx_list[af] if isinstance(pfx_list[af],bool) else str(pfx_list[af])
              for af in ('ipv4','ipv6') if af in pfx_list
      }
    if not link.prefix:
      link.pop('prefix',None)
    if pfx_list.get('unnumbered',None):
      link.unnumbered = True

  return pfx_list

def augment_lan_link(link: Box, addr_pools: Box, ndict: dict, defaults: Box) -> None:
  link_attr_base = get_link_base_attributes(defaults)
  if common.DEBUG:     # pragma: no cover (debugging)
    print(f'\nProcess LAN link {link}')

  pfx_list = augment_link_prefix(link,['lan'],addr_pools)
  interfaces = []

  if common.DEBUG:     # pragma: no cover (debugging)
    print(f'... on-link prefixes: {pfx_list}')

  link_cnt = 0
  for value in link[IFATTR]:
    node = value.node
    ifaddr = Box({},default_box=True)
    errmsg = get_node_link_address(
      node=ndict[node],
      ifdata=ifaddr,
      node_link_data=value,
      prefix=pfx_list,
      node_id=ndict[node].id)
    if errmsg:
      common.error(
        f'{errmsg}\n'+
        common.extra_data_printout(f'link data: {link}'),common.IncorrectValue,'links')

    ifaddr_add_module(ifaddr,link,defaults.get('module'))

    ifaddr = ifaddr + value
    ifaddr.pop('node',None)               # Remove the 'node' attribute from interface data -- now we know where it belongs
    link[IFATTR][link_cnt] = value

    if link.type != "stub":
      n_list = [ link[IFATTR][i].node for i in range(0,len(link[IFATTR])) if i != link_cnt ]
      ifaddr.name = link.get("name") or (node + " -> [" + ",".join(list(n_list))+"]")

    ifdata = interface_data(link=link,link_attr=link_attr_base,ifdata=ifaddr)
    node_intf = add_node_interface(ndict[node],ifdata,defaults)
    value.ifindex = node_intf.ifindex
    interfaces.append({ 'node': node, 'data':  node_intf })

    link_cnt = link_cnt + 1

  for node_if in interfaces:
    node_if['data'].neighbors = []
    for remote_if in interfaces:
      if remote_if['node'] != node_if['node'] or remote_if['data'].ifindex != node_if['data'].ifindex:
        ngh_data = { 'ifname': remote_if['data'].ifname, 'node': remote_if['node'] }
        for af in ('ipv4','ipv6'):
          if af in remote_if['data']:
            ngh_data[af] = remote_if['data'][af]
        node_if['data'].neighbors.append(ngh_data)

  if common.DEBUG:     # pragma: no cover (debugging)
    print(f'Final LAN link data: {link}\n')

def augment_p2p_link(link: Box, addr_pools: Box, ndict: dict, defaults: Box) -> typing.Optional[Box]:
  link_attr_base = get_link_base_attributes(defaults)
  if not defaults:      # pragma: no cover (almost impossible to get there)
    defaults = Box({})
  if common.DEBUG:     # pragma: no cover (debugging)
    print(f'\nProcess P2P link {link}')
  pfx_list = augment_link_prefix(link,['p2p','lan'],addr_pools)

  end_names = ['left','right']
  link_nodes: typing.List[Box] = []
  interfaces: typing.List[Box] = []

  if len(link[IFATTR]) > len(end_names): # pragma: no cover (this error is reported earlier)
    common.fatal(f"Internal error: Too many nodes specified on a P2P link {link}",'links')
    return None

  link[IFATTR] = sorted(link[IFATTR],key=lambda v: v.node)       # Keep nodes sorted in alphabetic order for historic reasons
  value: Box
  for intf_cnt,value in enumerate(link[IFATTR]):
    node = value.node
    ecount = len(link_nodes)
    ifaddr = Box({},default_box=True,box_dots=True)
    if link.get('unnumbered',None):
      ifaddr.unnumbered = True

    if not isinstance(value,dict):        # pragma: no cover -- caught earlier in adjust_interface_list
      common.error(f'Attributes for node {node} on link {link} must be a dictionary',common.IncorrectValue,'links')
      return None

    errmsg = get_node_link_address(
      node=ndict[node],
      ifdata=ifaddr,
      node_link_data=value,
      prefix=pfx_list,
      node_id=ecount+1)
    if errmsg:
      common.error(
        f'{errmsg}\n'+
        common.extra_data_printout(f'link data: {link}'),common.IncorrectValue,'links')

    ifaddr_add_module(ifaddr,link,defaults.get('module'))

    ifaddr = ifaddr + value
    ifaddr.pop('node',None)               # Remove the 'node' attribute from interface data -- now we know where it belongs
#    link[IFATTR][intf_cnt] = value
    link_nodes.append(Box({ 'name': node, 'link': value, 'ifaddr': ifaddr }))

    ifdata = interface_data(link=link,link_attr=link_attr_base,ifdata=ifaddr)
    if 'bridge' in link:
      ifdata.bridge = link.bridge
    ifdata.name = link.get("name") or (link[IFATTR][intf_cnt].node + " -> " + link[IFATTR][1-intf_cnt].node)
    dict_2_update = add_node_interface(ndict[node],ifdata,defaults)
    value.ifindex = dict_2_update.ifindex
    interfaces.append(dict_2_update)

  if not 'name' in link:
    link.name = link_nodes[0].name + " - " + link_nodes[1].name

  for i in range(0,2):
    link[end_names[i]] = { 'node': link_nodes[i]['name'],'ifname': interfaces[i].get('ifname') }

    remote = link_nodes[1-i].name
    interfaces[i]['neighbors'] = [{
        'node': remote,
        'ifname': interfaces[1-i]['ifname']
      }]
    for af in ('ipv4','ipv6'):
      if af in interfaces[1-i]:
        interfaces[i]['neighbors'][0][af] = interfaces[1-i][af]
      if af in interfaces[i]:
        link[end_names[i]][af] = interfaces[i][af]

  return link

def check_link_attributes(data: Box, nodes: dict, valid: set) -> bool:
  ok = True
  for k in data.keys():
    if k not in nodes and k not in valid:
      common.error("Invalid link attributes '%s' in %s" % (k,data),common.IncorrectValue,'links')
      ok = False

  return ok

def set_link_type_role(link: Box, pools: Box, nodes: Box) -> None:
  node_cnt = len(link[IFATTR])      # Set the number of attached nodes (used in many places further on)
  link['node_count'] = node_cnt

  host_count = 0                    # Count the number of hosts attached to the link
  for ifdata in link[IFATTR]:
    if nodes[ifdata.node].get('role','') == 'host':
      host_count = host_count + 1

  if host_count > 0:                # If we have hosts and a single router attached to a non-VLAN link, set link role to 'stub'
    link.host_count = host_count    # ... VLAN case will be set in the VLAN module
    if not 'role' in link and host_count == node_cnt - 1 and not 'vlan_name' in link:
      link.role = 'stub'

  if 'type' in link:                # Link type already set, nothing to do
    return

  link.type = 'lan' if node_cnt > 2 else 'p2p' if node_cnt == 2 else 'stub'     # Set link type based on number of attached nodes

  if host_count > 0:
    link.type = 'lan'

  return

def check_link_type(data: Box) -> bool:
  node_cnt = data.get('node_count') # link_node_count(data,nodes)
  link_type = data.get('type')

  if 'mtu' in data and not isinstance(data.mtu,int): # pragma: no cover
    common.error(f'MTU parameter should be an integer: {data}',common.IncorrectValue,'links')

  if not link_type: # pragma: no cover (shouldn't get here)
    common.fatal('Internal error: link type still undefined in check_link_type: %s' % data,'links')
    return False

  if node_cnt == 0:
    common.error('No valid nodes on link %s' % data,common.MissingValue,'links')
    return False

  if link_type == 'stub' and node_cnt > 1:
    common.error('More than one node connected to a stub link: %s' % data,common.IncorrectValue,'links')
    return False

  if link_type == 'p2p' and node_cnt != 2:
    common.error('Point-to-point link needs exactly two nodes: %s' % data,common.IncorrectValue,'links')
    return False

  if not link_type in [ 'stub','p2p','lan','vlan_member']:
    common.error('Invalid link type %s: %s' % (link_type,data),common.IncorrectValue,'links')
    return False
  return True

#
# Interface Feature Check -- validate that the selected addressing works on target lab devices
#

def interface_feature_check(nodes: Box, defaults: Box) -> None:
  for node,ndata in nodes.items():
    features = devices.get_device_features(ndata,defaults)
    for ifdata in ndata.get('interfaces',[]):
      if 'ipv4' in ifdata:
        if isinstance(ifdata.ipv4,bool) and ifdata.ipv4 and \
            not features.initial.ipv4.unnumbered:
          common.error(
            f'Device {ndata.device} does not support unnumbered IPv4 interfaces used on\n'+
            f'.. node {node} interface {ifdata.ifname} (link {ifdata.name})',
            common.IncorrectValue,
            'interfaces')
      if 'ipv6' in ifdata:
        if isinstance(ifdata.ipv6,bool) and ifdata.ipv6 and \
            not features.initial.ipv6.lla:
          common.error(
            f'Device {ndata.device} does not support LLA-only IPv6 interfaces used on\n'+
            f'.. node {node} interface {ifdata.ifname} (link {ifdata.name})',
            common.IncorrectValue,
            'interfaces')

def set_default_gateway(link: Box, nodes: Box) -> None:
  if not 'host_count' in link:      # No hosts attached to the link, get out
    return

  link.pop('host_count',None)
  if not 'gateway' in link:
    gateway = None
    for ifdata in link[IFATTR]:
      if nodes[ifdata.node].get('role','') != 'host' and 'ipv4' in ifdata:
        link.gateway.ipv4 = ifdata.ipv4
        break
  else:
    if not isinstance(ifdata.gateway,dict) or not 'ipv4' in ifdata.gateway:  # pragma: no cover
      common.error(
        f'Gateway attribute specified on {link} is not a dictionary with ipv4 key',
        common.IncorrectValue,
        'links')

  if not 'gateway' in link:         # Didn't find a usable gateway, exit
    return

  for ifdata in link[IFATTR]:                             # Copy link gateway to all hosts attached to the link
    if nodes[ifdata.node].get('role','') == 'host':       # Set gateway only for hosts
      for interface in nodes[ifdata.node].interfaces:     # Find the corresponding host interface
        if link.linkindex == interface.linkindex:
          if interface.ifindex == 1:                      # Set the default gateway only on the first host interface
            interface.gateway = link.gateway

"""
Set node.af flags to indicate that the node has IPv4 and/or IPv6 address family configured
"""
def set_node_af(nodes: Box) -> None:
  for n in nodes.values():
    for af in ['ipv4','ipv6']:
      if af in n.get('loopback',{}):
        n.af[af] = True
        continue

      for l in n.get('interfaces',[]):
        if af in l:
          n.af[af] = True
          continue

def set_linkindex(topology: Box) -> None:
  if not 'links' in topology:
    return

  linkindex = topology.defaults.get('link_index',1)
  for link in topology.links:
    link.linkindex = linkindex
    linkindex = linkindex + 1

def transform(link_list: typing.Optional[Box], defaults: Box, nodes: Box, pools: Box) -> typing.Optional[Box]:
  if not link_list:
    return None

  link_attr_full = get_link_full_attributes(defaults)

  for link in link_list:
    if not check_link_attributes(data=link,nodes=nodes,valid=link_attr_full):
      continue

    set_link_type_role(link=link,pools=pools,nodes=nodes)
    if not check_link_type(data=link):
      continue

    if link.type == 'p2p':
      augment_p2p_link(link,pools,nodes,defaults=defaults)
    else:
      if not 'bridge' in link:
        link['bridge'] = "%s_%d" % (defaults.name[0:10],link.linkindex) # max 15 chars on Linux
      elif len(link['bridge']) > 15:
        common.error(
            f'Bridge name {link["bridge"]} has more than 15 characters',
            common.IncorrectValue,
            'interfaces')
      augment_lan_link(link,pools,nodes,defaults=defaults)

    set_default_gateway(link,nodes)

  interface_feature_check(nodes,defaults)
  set_node_af(nodes)
  return link_list



'''
plugin - implement custom topology transformation plugins
'''

import os
import sys
import typing

from box import Box
from importlib import import_module
from .. import common

def load_plugin_from_path(path: str, plugin: str) -> typing.Optional[object]:
  module_path = path+'/'+plugin
  module_name = module_path
  module: typing.Optional[object] = None
  is_package = os.path.isdir(module_path)
  config_name = None

  if is_package:
    sys.path = [ module_path ] + sys.path
    module_name = module_path + '/plugin.py'
    config_name = plugin
    plugin = 'plugin'
  else:
    sys.path = [ path ] + sys.path
    if os.path.exists(module_name+'.py'):
      module_name = module_name + '.py'

  if os.path.exists(module_name):
    try:
      module = import_module(plugin)
    except:
      print("Cannot load plugin %s\n%s" % (module_name,str(sys.exc_info()[1])))
      common.fatal('Aborting the transformation process','plugin')

    if is_package:
      setattr(module,'config_name',config_name)
  else:
    module = None

  sys.path = sys.path[1:]
  if common.VERBOSE:
    print("loaded plugin %s" % module)
  return module

def init(topology: Box) -> None:
  if not 'plugin' in topology:
    return

  topology.Plugin = []
  for pname in topology.plugin:
    plugin = None
    search_path = ('.',common.netsim_package_path+'/extra',pname)
    for path in search_path:
      if not plugin:
        plugin = load_plugin_from_path(path,pname)
    if plugin:
      topology.Plugin.append(plugin)
    else:
      common.error(f"Cannot find plugin {pname} in {search_path}",common.IncorrectValue,'plugin')

def execute(action: str, topology: Box) -> None:
  if not 'Plugin' in topology:
    return

  for plugin in topology.Plugin:
  	if hasattr(plugin,action):
  	  func = getattr(plugin,action)
  	  func(topology)



#
# MPLS (LDP/BGP-LU) transformation module
#
import typing

from box import Box

from . import _Module,_routing
from .. import common
from ..common import AF_LIST,BGP_SESSIONS
from .. import data
from ..augment import devices

DEFAULT_BGP_LU: dict = {
  'ipv4': [ 'ibgp','ebgp' ],
  'ipv6': [ 'ibgp','ebgp' ]
}

DEFAULT_VPN_AF: dict = {
  'ipv4': [ 'ibgp' ],
  'ipv6': [ 'ibgp' ]
}

FEATURE_NAME: dict = {
  'ldp': 'LDP-based label distribution',
  'bgp': 'BGP Labeled Unicast',
  'vpn': 'MPLS/VPN'
}

def node_adjust_ldp(node: Box, topology: Box, features: Box) -> None:
  if not 'ipv4' in node.get('af',{}):
    common.error(
      f'You cannot enable MPLS LDP on node {node.name} without IPv4 address family',
      common.MissingValue,
      'mpls')
    return

  node.ldp = node.mpls.ldp
  node.mpls.pop('ldp',None)
  _routing.router_id(node,'ldp',topology.pools)

  for intf in node.get('interfaces',[]):
    intf_ldp = data.get_from_box(intf,'mpls.ldp')
    if 'vrf' in intf:
      if not intf_ldp:
        continue
      if not features.mpls.csc:
        common.error(
          f'Device {node.device} does not support MPLS CsC (LDP in VRF) configured on {node.name}',
          common.IncorrectValue,
          'mpls')
        continue

    if not intf_ldp is None and 'mpls' in intf:
      data.bool_to_defaults(intf.mpls,'ldp')
      if 'ldp' in intf.mpls:
        intf.ldp = intf.mpls.ldp

    if not _routing.external(intf,'ldp'):
      _routing.passive(intf,'ldp')

  _routing.remove_unused_igp(node,'ldp')

def validate_mpls_bgp_parameter(node: Box, feature: str) -> bool:
  if isinstance(node.mpls[feature],list):
    session_list = node.mpls[feature]
    if not data.validate_list_elements(session_list,BGP_SESSIONS,f'nodes.{node.name}.mpls.{feature}'):
      common.error(
        f'Invalid BGP session type in nodes.{node.name}.mpls.{feature} parameter',
        common.IncorrectValue,
        'mpls')
      return False

    node.mpls[feature] = Box({})
    for af in node.af:
      node.mpls[feature][af] = session_list
  elif isinstance(node.mpls[feature],Box):
    for af in AF_LIST:
      if not af in node.mpls[feature]:
        continue

      if data.must_be_list(node.mpls[feature],af,f'nodes.{node.name}.mpls.{feature}') is None:
        return False

      if not data.validate_list_elements(node.mpls[feature][af],BGP_SESSIONS,f'nodes.{node.name}.mpls.{feature}.{af}'):
        common.error(
          f'Invalid BGP session type in nodes.{node.name}.mpls.{feature}.{af} parameter',
          common.IncorrectValue,
          'mpls')
        return False
  else:
    common.error(
      f'nodes.{node.name}.mpls.{feature} parameter must be a boolean, list, or dictionary',
      common.IncorrectValue,
      'mpls')
    return False

  return True

def node_adjust_bgplu(node: Box, topology: Box, features: Box) -> None:
  if not validate_mpls_bgp_parameter(node,'bgp'):
    return

  if not 'neighbors' in node.get('bgp',{}):
    return

  for n in node.bgp.neighbors:
    for af in ['ipv4','ipv6']:
      if af in n and af in node.mpls.bgp and n.type in node.mpls.bgp[af]:
        n[af+'_label'] = True

def prune_mplsvpn_af(setting: Box, node: Box) -> None:
  vrf_af :dict = {}

  for af in AF_LIST:
    for vdata in node.get('vrfs',{}).values():
      if af in vdata.af:
        vrf_af[af] = True
        break

  for af in AF_LIST:
    if af in setting and not af in vrf_af and not 'ebgp' in setting[af]:
      setting.pop(af,None)

def node_adjust_mplsvpn(node: Box, topology: Box, features: Box) -> None:
  if not validate_mpls_bgp_parameter(node,'vpn'):
    return

  if not node.bgp.get('rr',False):
    prune_mplsvpn_af(node.mpls.vpn,node)

  for n in node.bgp.neighbors:
    if 'ipv4' in n:
      for af in AF_LIST:
        if af in node.mpls.vpn:
          n['vpn'+af.replace('ip','')] = n.ipv4

class MPLS(_Module):

  def node_pre_transform(self, node: Box, topology: Box) -> None:
    global DEFAULT_BGP_LU
    if not 'mpls' in node:
      return

    data.bool_to_defaults(node.mpls,'ldp',{})
    if 'ldp' in node.mpls:
      data.must_be_bool(node.mpls.ldp,'disable_unlabeled',f'nodes.{node.name}.mpls.ldp')
      if not any(m in ['ospf','isis','eigrp'] for m in node.module):
        common.error(
          f'You cannot enable LDP on node {node.name} without an IGP',
          common.MissingValue,
          'mpls')

    data.bool_to_defaults(node.mpls,'bgp',DEFAULT_BGP_LU)
    data.bool_to_defaults(node.mpls,'vpn',DEFAULT_VPN_AF)
    if 'bgp' in node.mpls or 'vpn' in node.mpls:
      if not 'bgp' in node.module:
        common.error(
          f'You cannot enable BGP-LU or MPLS/VPN on node {node.name} without BGP module',
          common.MissingValue,
          'mpls')

  def node_post_transform(self, node: Box, topology: Box) -> None:
    global FEATURE_NAME

    if not node.mpls:     
      return

    features = devices.get_device_features(node,topology.defaults)

    for fn in FEATURE_NAME.keys():
      if fn in node.mpls and not features.mpls[fn]:
        common.error(
          f'Device {node.device} used by {node.name} does not support {FEATURE_NAME[fn]}',
          common.IncorrectValue,
          'mpls')

    if 'ldp' in node.mpls:
      node_adjust_ldp(node,topology,features)

    if 'bgp' in node.mpls:
      node_adjust_bgplu(node,topology,features)

    if 'vpn' in node.mpls:
      node_adjust_mplsvpn(node,topology,features)



#
# Vagrant/libvirt provider module
#
from box import Box

from . import _TopologyOutput
from .. import providers
from .. import common

class ProviderConfiguration(_TopologyOutput):

  def write(self, topology: Box) -> None:
    provider = providers._Provider.load(topology.provider,topology.defaults.providers[topology.provider])
    provider.call('pre_output_transform',topology)

    filename = None
    if hasattr(self,'filenames'):
      filename = self.filenames[0]
      if len(self.filenames) > 1:
        common.error('Extra output filename(s) ignored: %s' % str(self.filenames[1:]),common.IncorrectValue,'provider')

    if self.format:
      common.error('Specified output format(s) %s ignored' % self.format,common.IncorrectValue,'provider')

    provider.create(topology,filename)



#
# netlab down command
#
# * Transform lab topology or read transformed lab topology from snapshot file
# * Stop the lab, including provider-specific pre- and post-stop hooks
# * Clean up the working directory (optional)
#
import argparse
import typing
import textwrap
from box import Box

from . import common_parse_args, topology_parse_args, load_topology, load_snapshot_or_topology, external_commands,fs_cleanup
from .. import read_topology,augment,common
from .. import providers

#
# CLI parser for 'netlab down' command
#
def down_parse(args: typing.List[str]) -> argparse.Namespace:
  parser = argparse.ArgumentParser(
    parents=[ topology_parse_args() ],
    prog="netlab down",
    description='Destroy the virtual lab')

  parser.add_argument(
    '-v','--verbose',
    dest='verbose',
    action='count',
    default = 0,
    help='Verbose logging (where applicable)')
  parser.add_argument(
    '--cleanup',
    dest='cleanup',
    action='store_true',
    help='Remove all configuration files created by netlab create')
  parser.add_argument(
    '--snapshot',
    dest='snapshot',
    action='store',
    nargs='?',
    const='netlab.snapshot.yml',
    help='Transformed topology snapshot file')
  parser.add_argument(
    dest='topology', action='store', nargs='?',
    help='Topology file (default: topology.yml)')

  return parser.parse_args(args)

def down_cleanup(topology: Box, verbose: bool = False) -> None:
  cleanup_list = topology.defaults.providers[topology.provider].cleanup or []
  cleanup_list.extend(topology.defaults.automation.ansible.cleanup)
  cleanup_list.append('netlab.snapshot.yml')
  fs_cleanup(cleanup_list,verbose)

def run(cli_args: typing.List[str]) -> None:
  args = down_parse(cli_args)
  topology = load_snapshot_or_topology(args)

  if args.topology:
    print(f"Reading lab topology from {args.topology}")
  else:
    print(f"Reading transformed lab topology from snapshot file {args.snapshot}")

  if topology is None:
    common.fatal('... could not read the lab topology, aborting')
    return

  settings = topology.defaults
  external_commands.run_probes(settings,topology.provider,1)

  provider = providers._Provider.load(topology.provider,topology.defaults.providers[topology.provider])
  provider.call('pre_output_transform',topology)

  if hasattr(provider,'pre_stop_lab') and callable(provider.pre_stop_lab):
    provider.pre_stop_lab(topology)

  external_commands.stop_lab(settings,topology.provider,2,"netlab down")

  if hasattr(provider,'post_stop_lab') and callable(provider.post_stop_lab):
    provider.post_stop_lab(topology)

  if args.cleanup:
    external_commands.print_step(3,"Cleanup configuration files",spacing = True)
    down_cleanup(topology,True)



#!/usr/bin/env python3
#
# Print netlab usage
#

import sys
import typing

try:
  from importlib import resources
except ImportError:
  import importlib_resources as resources # type: ignore

def print_usage(fname: str) -> None:
  package = '.'.join(__name__.split('.')[:-1])
  with resources.open_text(package,fname) as fid:
    print(fid.read())

def run(args: typing.List[str]) -> None:
  print_usage('usage.txt')



'''
Topology-level transformation:

* Check for required elements (nodes, defaults)
* Check for extraneous elements
* Adjust 'provider' parameter
* Create expanded topology file in YAML format (mostly for troubleshooting purposes)
'''

import os
from box import Box

from .. import common
from .. import data

#
# Extend link/node/global attribute lists with extra attributes
#
def extend_attribute_list(settings: Box, attribute_path: str = 'topology.defaults', always_valid: list = []) -> None:
  if not 'extra_attributes' in settings:
    return

  for k in settings.extra_attributes.keys():           # Iterate over extensions
    if not k in settings.get('attributes',{}):         # Check that the extension is valid
      if k in always_valid:                            # ... some extensions are always valid (needed for modules)
        settings.attributes[k] = []                    # ... in which case we have to start with an empty list
      else:                                            # ... for everything else, throw an error
        common.error(
          f'Invalid extra_attribute {k} -- not present in configurable {attribute_path} attributes',
          common.IncorrectValue,
          'topology')

    data.must_be_list(                               # Make sure the extension is a list so it's safe to iterate over
      parent = settings.extra_attributes,
      key = k,
      path = f'{attribute_path}.extra_attributes.{k}')

    common.exit_on_error()

    for v in settings.extra_attributes[k]:             # Have to iterate over values in the custom attribute list
      if not v in settings.attributes[k]:              # ... to prevent duplicate values in attribute lists
        settings.attributes[k].append(v)               # Going through sets is not an option due to element rearrangements

#
# Extend attribute lists for all top-level elements of the defaults dictionary
# with 'attributes' and 'extra_attributes' keys
#
def extend_module_attribute_list(topology: Box) -> None:
  for k in topology.defaults.keys():
    if isinstance(topology.defaults[k],dict):
      if 'extra_attributes' in topology.defaults[k]:
        if not 'attributes' in topology.defaults[k]:   # pragma: no cover (things would break way before this point)
          topology.defaults[k].attributes = {}
        extend_attribute_list(topology.defaults[k],f'topology.defaults.{k}',['global','node','link'])

def check_required_elements(topology: Box) -> None:
  invalid_topo = False
  for rq in ['nodes']:
    if not rq in topology:
      common.error(f"Lab topology is missing mandatory {rq} element",category=common.MissingValue,module="topology")
      invalid_topo = True
    elif not topology.get(rq):
      common.error(f"Required topology element {rq} is empty",category=common.MissingValue,module="topology")
      invalid_topo = True

  if invalid_topo:
    common.fatal("Fatal topology errors, aborting")

  if not 'name' in topology:
    topo_name = os.path.basename(os.path.dirname(os.path.realpath(topology['input'][0])))
    topology.name = topo_name

  if 'module' in topology:
    data.must_be_list(topology,'module','')
    topology.defaults.module = topology.module

  topology.defaults.name = topology.name

def check_global_elements(topology: Box) -> None:
  topo_elements = topology.defaults.attributes.get('global',[]) + topology.defaults.attributes.get('internal',[])
  module_list   = [ modname for modname in topology.defaults.keys() if 'supported_on' in topology.defaults[modname] ]
  extra_modattr: dict = {}
  for m,x_list in [ (modname,topology.defaults[modname].attributes.extra['global']) \
                    for modname in topology.defaults.keys() \
                      if data.get_from_box(topology.defaults[modname],'attributes.extra.global') ]:
    for attr in x_list:
      extra_modattr[attr] = m

  if topology.get('module'):
    topo_elements = topo_elements + topology.module

  for k in topology.keys():
    if not k in topo_elements:
      if k in module_list:
        common.error(
          f'Add module {k} to global- or node configuration modules to use {k}.something global attribute',
          category=common.IncorrectValue,
          module="topology")
      elif k in extra_modattr:
        common.error(
          f'Global attribute {k} is defined by module {extra_modattr[k]} which is not used in your topology',
          category=common.IncorrectValue,
          module="topology")
      else:
        common.error(
          f'Unknown top-level element {k} -- typo?',
          category=common.IncorrectValue,
          module="topology")

#
# Find virtualization provider, set provider and defaults.provider to that value
#
# Note: defaults.provider is needed in some output routines that get defaults data structure
# but not the whole topology
#
def adjust_global_parameters(topology: Box) -> None:
  if not 'provider' in topology:
    topology.provider = topology.defaults.provider
  else:
    topology.defaults.provider = topology.provider

  if not topology.provider:
    common.fatal('Virtualization provider is not defined in either "provider" or "defaults.provider" elements')

  providers = topology.defaults.providers
  if not topology.provider in providers:
    plist = ', '.join(providers.keys())
    common.fatal('Unknown virtualization provider %s. Supported providers are: %s' % (topology.provider,plist))

  # Adjust defaults with provider-specific defaults
  #
  for k in ['addressing']:
    if k in topology.defaults.providers[topology.provider]:
      topology.defaults[k] = topology.defaults[k] + topology.defaults.providers[topology.provider][k]

#
# Cleanup the topology
#

def cleanup_topology(topology: Box) -> Box:
  topo_copy = Box(topology)

  # Remove PFX generators from addressing section
  #
  for k,v in topo_copy.addressing.items():
    for p in list(v.keys()):
      if "_pfx" in p or "_eui" in p:
        v.pop(p,None)

  return topo_copy



#
# netlab initial command
#
# Deploys initial device configurations
#
import typing
import os
import argparse

from . import common_parse_args
from . import ansible

#
# CLI parser for 'netlab initial' command
#
def initial_config_parse(args: typing.List[str]) -> typing.Tuple[argparse.Namespace, typing.List[str]]:
  parser = argparse.ArgumentParser(
    parents=[ common_parse_args() ],
    prog="netlab initial",
    description='Initial device configurations',
    epilog='All other arguments are passed directly to ansible-playbook')

  parser.add_argument(
    '-i','--initial',
    dest='initial', action='store_true',
    help='Deploy just the initial configuration')
  parser.add_argument(
    '-m','--module',
    dest='module', action='store',nargs='?',const='*',
    help='Deploy module-specific configuration (optionally including a list of modules separated by commas)')
  parser.add_argument(
    '-c','--custom',
    dest='custom', action='store_true',
    help='Deploy custom configuration templates (specified in "config" group or node attribute)')
  parser.add_argument(
    '--fast',
    dest='fast', action='store_true',
    help='Use "free" strategy in Ansible playbook for faster configuration deployment')
  parser.add_argument(
    '-o','--output',
    dest='output', action='store',nargs='?',const='config',
    help='Create a directory with initial configurations instead of deploying them (default output directory: config)')
  return parser.parse_known_args(args)

def run(cli_args: typing.List[str]) -> None:
  (args,rest) = initial_config_parse(cli_args)

  if args.verbose:
    rest = ['-' + 'v' * args.verbose] + rest

  if args.initial:
    rest = ['-t','initial'] + rest

  if args.quiet:
    os.environ["ANSIBLE_STDOUT_CALLBACK"] = "selective"

  if args.module:
    if args.module != "*":
      rest = ['-e','modlist='+args.module] + rest
    rest = ['-t','module'] + rest
  
  if args.custom:
    rest = ['-t','custom'] + rest

  if args.output:
    rest = ['-e','config='+os.path.abspath(args.output) ]

  if args.fast or os.environ.get('NETSIM_FAST_CONFIG',None):
    rest = ['-e','netsim_strategy=free']

  if args.logging or args.verbose:
    print("Ansible playbook args: %s" % rest)

  if args.output:
    ansible.playbook('create-config.ansible',rest)
    print("\nInitial configurations have been created in the %s directory" % args.output)
  else:
    ansible.playbook('initial-config.ansible',rest)



#
# Vagrant/libvirt provider module
#

import subprocess
import re
from box import Box
import pathlib

from .. import common
from . import _Provider

LIBVIRT_MANAGEMENT_NETWORK_NAME = "vagrant-libvirt"
LIBVIRT_MANAGEMENT_NETWORK_FILE = "templates/provider/libvirt/vagrant-libvirt.xml"

class Libvirt(_Provider):

  def transform_node_images(self, topology: Box) -> None:
    self.node_image_version(topology)

  def pre_output_transform(self, topology: Box) -> None:
    for node in topology.nodes.values():
      for intf in node.interfaces:
        if intf.get('linkindex') and not intf.get('virtual_interface'):
          link = topology.links[intf.linkindex - 1]
          if len(link.interfaces) == 2:
            intf.libvirt.type = "tunnel"
            link.pop("bridge",None)
            remote_if_list = [ rif for rif in link.interfaces if rif.node != node.name or rif.ifindex != intf.ifindex ]
            if len(remote_if_list) != 1:
              common.fatal(
                f'Cannot find remote interface for P2P link\n... node {node.name}\n... intf {intf}\n... link {link}\n... iflist {remote_if_list}')
              return

            remote_if = remote_if_list[0]
            intf.remote_ifindex = remote_if.ifindex
            intf.remote_id = topology.nodes[remote_if.node].id
            if not intf.remote_id:
              common.fatal(
                f'Cannot find remote node ID on a P2P link\n... node {node.name}\n... intf {intf}\n... link {link}')
              return

  def pre_start_lab(self, topology: Box) -> None:
    common.print_verbose('pre-start hook for libvirt')
    # Starting from vagrant-libvirt 0.7.0, the destroy actions deletes all the networking
    #  including the "vagrant-libvirt" management network.
    #  Let's re-create it if missing!
    try:
      result = subprocess.run(['virsh','net-info',LIBVIRT_MANAGEMENT_NETWORK_NAME],capture_output=True,text=True)
      if result.returncode == 1:
        # When ret code is 1, the network is missing
        common.print_verbose('creating missing %s network' % LIBVIRT_MANAGEMENT_NETWORK_NAME)
        net_template_xml = pathlib.Path(__file__).parent.parent.joinpath(LIBVIRT_MANAGEMENT_NETWORK_FILE).resolve()
        result2 = subprocess.run(['virsh','net-define',net_template_xml],capture_output=True,check=True,text=True)
    except subprocess.CalledProcessError as e:
      common.error('Exception in net handling for libvirt network %s: [%s] %s' % (LIBVIRT_MANAGEMENT_NETWORK_NAME, e.returncode, e.stderr), module='libvirt')
    return

  def post_start_lab(self, topology: Box) -> None:
    common.print_verbose('libvirt lab has started, fixing Linux bridges')
    for l in topology.links:
      brname = l.get('bridge',None)
      if not brname:
        continue
      try:
        print(l)
        result = subprocess.run(['virsh','net-info',brname],capture_output=True,check=True,text=True)
      except:
        common.error('Cannot run net-info for libvirt network %s' % brname, module='libvirt')
        continue

      match = None
      if result and result.returncode == 0:
        match = re.search("Bridge:\\s+(.*)$",result.stdout,flags=re.MULTILINE)

      if match:
        common.print_verbose("... network %s maps into %s" % (brname,match.group(1)))
        subprocess.run(['sudo','sh','-c','echo 0x4000 >/sys/class/net/%s/bridge/group_fwd_mask' % match.group(1)],check=True)
        common.print_verbose("... setting LLDP enabled flag on %s" % (match.group(1)))
      else:
        common.error('Cannot get Linux bridge name for libvirt network %s' % brname, module='libvirt')



#
# Dynamic module transformation framework
#
# Individual configuration modules are defined as Python files within this directory inheriting
# Module class and adding transformation methods
#
import typing

from box import Box

# Related modules
from .. import common
from ..callback import Callback
from ..augment.links import IFATTR
from ..augment import devices

# List of attributes we don't want propagated from defaults to global/node
#
no_propagate_list = ["attributes","extra_attributes","requires","supported_on","no_propagate","config_after","transform_after"]

class _Module(Callback):

  def __init__(self, data: Box) -> None:
    pass

  @classmethod
  def load(self, module: str, data: Box) -> typing.Any:
    module_name = __name__+"."+module
    obj = self.find_class(module_name)
    if obj:
      return obj(data)
    else:
      return _Module(data)

"""
pre_transform: executed just before the main data model transformation is started

* Adjust global and node data structures
* Call module-specific node transformation code
* Call module-specific link transformation code
"""
def pre_transform(topology: Box) -> None:
  adjust_modules(topology)
  common.exit_on_error()

  check_module_parameters(topology)
  common.exit_on_error()

  module_transform("pre_transform",topology)
  node_transform("pre_transform",topology)
  link_transform("pre_transform",topology)

"""
pre_node_transform: executed just before the node data model transformation is started

"""
def pre_node_transform(topology: Box) -> None:
  module_transform("pre_node_transform",topology)
  node_transform("pre_node_transform",topology)
  link_transform("pre_node_transform",topology)

"""
pre_link_transform: executed just before the link data model transformation is started

"""
def pre_link_transform(topology: Box) -> None:
  module_transform("pre_link_transform",topology)
  node_transform("pre_link_transform",topology)
  link_transform("pre_link_transform",topology)

"""
post_transform:
  execute module-specific code after the main link- and node
  transformations has completed
"""
def post_transform(topology: Box) -> None:
  check_supported_node_devices(topology)       # A bit late, but we can do this check only after node data has been adjusted
  common.exit_on_error()
  copy_node_data_into_interfaces(topology)     # Copy node attributes that match interface attributes into interfaces
  module_transform("post_transform",topology)
  node_transform("post_transform",topology)
  link_transform("post_transform",topology)
  reorder_node_modules(topology)               # Make sure modules are configured in dependency order (#86)

# Set default list of modules for nodes without specific module list
#
def augment_node_module(topology: Box) -> None:
  if not 'module' in topology:
    return

  module = topology['module']
  for name,n in topology.nodes.items():
    if not 'module' in n and n.get('role') != 'host' and devices.get_device_attribute(n,'role',topology.defaults) != 'host':
      n.module = module

# Check whether the modules defined on individual nodes are valid
# and supported
#
def check_supported_node_devices(topology: Box) -> None:
  for name,n in topology.nodes.items():
    for m in n.get("module",[]):                                # Iterate across all modules used by a node
      if not m in topology.defaults:                            # Do we know about the module?
        common.error(
          f"Unknown module {m} used by node {name}",
          common.IncorrectValue,
          'modules')
        continue
      mod_def = topology.defaults[m]                            # Get module defaults
      if mod_def and "supported_on" in mod_def :                # Are they sane and do they include supported device list?
        if not n.device in topology.defaults[m].supported_on:   # ... and is the device on the list?
          common.error(
            f"Device type {n.device} used by node {name} is not supported by module {m}",
            common.IncorrectValue,
            'modules')
          continue

# Merge global module parameters with per-node module parameters
#
# Remove "no_propagate" values (default: "attributes")
# before merging the global settings
#

def merge_node_module_params(topology: Box) -> None:
  for name,n in topology.nodes.items():
    if 'module' in n:
      for m in n.module:
        if not m in topology.defaults:      # Cannot propagate parameters of unknown module
          continue                          # ... error will be reported later
        if m in topology:
          global_settings = get_propagated_global_module_params(m,topology.get(m,{}),topology.defaults[m])
          if global_settings:
            n[m] = global_settings + n[m]

        dev_settings = devices.get_device_attribute(n,m,topology.defaults)
        if dev_settings:
          n[m] = get_propagated_global_module_params(m,dev_settings,topology.defaults[m]) + n[m]

        if isinstance(topology.defaults[m],dict):
          default_settings = get_propagated_global_module_params(m,topology.defaults[m],topology.defaults[m])
          if default_settings:
            n[m] = default_settings + n[m]

# Get propagated global parameters from settings (top-level or device-level)
#
# settings - dictionary with module settings
# mod_settings - default module settings
#
def get_propagated_global_module_params(module: str, settings: Box,mod_settings: Box) -> Box:
  global no_propagate_list

  global_copy = Box(settings)                  # Make a fresh copy of the settings
  no_propagate = list(no_propagate_list)       # ... and global no_propagate list
  if "no_propagate" in mod_settings:
    no_propagate.extend(mod_settings.no_propagate)
  for remove_key in no_propagate:
    global_copy.pop(remove_key,None)
  return global_copy

'''
adjust_global_modules: last phase of global module adjustments

* add node-specific modules into global list of modules after the node
  modules have been set to default global values
* merge default settings with global settings
* copy global settings into node settings (apart from no_propagate attributes)
'''
def adjust_global_modules(topology: Box) -> None:
  mod_dict = { m : None for m in topology.get('module',[]) }
  for name,n in topology.nodes.items():
    for m in n.get('module',[]):
      mod_dict[m] = None

  if not mod_dict:
    return

  """
  Merge default module parameters with global module parameters

  A caveat: don't merge key specified in defaults "no_propagate" list --
  forces us to do a deep copy of global parameters, and then eliminate
  the ones we don't want.

  We couldn't just iterate because we need a deep merge, and we can't remove
  the no_propagate parameters from the global settings because they might be
  needed in further transformation code.
  """
  global no_propagate_list

  topology.module = list(mod_dict.keys())
  for m in topology.module:                                     # Iterate over all configured modules
    if not m in topology.defaults:                              # Does this module have defaults?
      common.error(
        f"Unknown module {m} (we found no system defaults for it)",
        common.IncorrectValue,
        'module')
      continue                                                  # Nope. Weird, but doesn't matter right now.
    mod_def = topology.defaults[m]
    if not isinstance(mod_def,dict):                               # Are module defaults a dict?
      common.fatal("Defaults for module %s should be a dict" % m)  # Nope? Too bad, crash right now, we can't live like that...

    default_copy = Box(mod_def)                                 # Got module defaults. Now copy them (we're gonna clobber them)
    no_propagate = list(no_propagate_list)                      # Always remove these default attributes (and make a fresh copy of the list)
    no_propagate.extend(default_copy.get("no_propagate", []))   # ... and whatever the module wishes to be removed
    for remove_key in no_propagate:                             # We got the list of unwanted attributes.
      default_copy.pop(remove_key,None)                         # ... remove them with extreme prejudice

    if len(default_copy):                        # Anything left? Let's merge it with existing settings
      topology[m] = default_copy + topology[m]   # Have to use this convoluted way to prevent generating empty dict

  reorder_node_modules(topology,'transform_after')

'''
add_module_extra_parameters: add extra module keywords (ex: 'vrfs' for 'vrf' module) to the list of attributes
'''
def add_module_extra_parameters(topology: Box) -> None:
  if not 'module' in topology:
    return

  for m in topology.module:                                     # Iterate through the global list of modules
    if 'extra' in topology.defaults[m].attributes:              # Does the module have 'extra' parameters?
      for k in topology.defaults[m].attributes.extra.keys():    # ... oh, it does, iterate through its keys (attribute levels)
        for attr in topology.defaults[m].attributes.extra[k]:   # Take every attribute from the list of extra attributes
          if not attr in topology.defaults.attributes[k]:       # ... and if it's not already in the global list of attributes
            topology.defaults.attributes[k].append(attr)        # ... append it to the global list

'''
adjust_modules: somewhat intricate multi-step config module adjustments

* Set node default modules based on global modules
* Adjust global module list based on node modules + copy default settings into topology settings
* Check whether the module parameters specified globally, or on node/link level, are valid
* Merge global module parametres into nodes
'''
def adjust_modules(topology: Box) -> None:
  augment_node_module(topology)
  merge_node_module_params(topology)
  adjust_global_modules(topology)
  add_module_extra_parameters(topology)
  module_transform("init",topology)
  check_module_parameters(topology)
  check_module_dependencies(topology)
  module_transform("pre_default",topology)
  node_transform("pre_default",topology)
  link_transform("pre_default",topology)
  if 'module' in topology:
    topology.defaults.module = topology.module

"""
check_module_parameters:

Verify global, node, link, and node-on-link data for all modules
that include _attributes_ element in default settings

parse_module_attributes:

Accept list or dict as module attributes. List format applies to
all levels (global, node, link, node-on-link), dict format specifies
lists for every level.

Global attributes are used as default value for node attributes.
Link attributes are used as default value for node-on-link attributes.
"""

def check_module_parameters(topology: Box) -> None:
  mod_attr = Box({},default_box=True,box_dots=True)

  for m in topology.get("module",[]):                      # Iterate over all active modules
    mod_def = topology.defaults.get(m,{})                  # Get module defaults
    if mod_def:                                            # Did we get something meaningful?
      if "attributes" in topology.defaults.get(m,{}):      # ... and does it include "attributes"?
        mod_attr[m] = parse_module_attributes(topology.defaults[m].attributes)

        if topology.get(m,{}):                             # Now we can start: are there global module parameters in topology?
          for k in topology[m].keys():                     # Got them - iterate over them
            if not k in mod_attr[m]["global"]:             # Did we get a parameter that is not in global attributes? Jeez... barf
              common.error(
                "Invalid global %s attribute %s" % (m,k),
                common.IncorrectValue,
                'module')

  for name,n in topology.nodes.items():  # Inspect all nodes
    for m in n.get("module",[]):         # Iterate over all node modules
      if mod_attr[m] and m in n:         # Does the current module have a list of attributes?
                                         # ...Does node have module attribute?
        for k in n[m].keys():            # Iterate over node-level module-specific attributes
          if not k in mod_attr[m].node:  # If the name of an attribute is not in the list of allowed
                                         # ...node-level attributes report error
            common.error(
              f"Node {name}: invalid attribute {k} for module {m}",
              common.IncorrectValue,
              'module')

  for g in topology.get('groups',{}):                    # Inspect node_data in groups
    if 'node_data' in topology.groups[g]:
      for m in topology.get('module',[]):                # Iterate over global modules
        if m in topology.groups[g].node_data:            # Does the group node_data contain module attributes?
          for k in topology.groups[g].node_data[m]:      # Iterate over module-specific attributes in node_data
            if not k in mod_attr[m].node:                # Is the attribute a valid node attribute for the module?
              common.error(
                f"node_data in group {g} contains invalid attribute {k} for module {m}",
                common.IncorrectValue,
                'groups')

  for l in topology.get("links",[]):        # Inspect all links - use get to avoid instantiating links key
    for m in topology.get("module",[]):     # Iterate over all known modules - get avoids instantiating module key
      if m in l and mod_attr[m] and l[m]:   # ... focusing on modules that have attributes specified on this link
                                            # ... and want to have their attributes checked

        # Module attribute could be a list of keys or a class type
        #
        if isinstance(l[m],Box) and isinstance(mod_attr[m].link,list):
          for k in l[m].keys():             # Iterate over link-level module attributes
            if not k in mod_attr[m].link:   # If the name of an attribute is not in the list of allowed
                                            # ... link-level attributes, report an error
              common.error(
                f"Invalid attribute {k} for module {m} on link {l}",
                common.IncorrectValue,
                'modules')
        else:
          if type(l[m]).__name__ != str(mod_attr[m].link):
            common.error(
              f"Invalid value type for attribute {k} for module {m} on link {l}, expected {mod_attr[m].link}",
              common.IncorrectValue,
              'modules')


    for intf in l.interfaces:                     # Iterate over all interfaces attached to the link
      n = intf.node                               # ... get node name
      node = topology.nodes[n]                    # ... and node data structure (so the expressions don't get too crazy)
      for m in node.get("module",[]):             # Iterate over all node modules
        if mod_attr[m] and m in intf:             # Does the current module have a list of attributes?
                                                  # ... and interface contains module attributes?
          # Module attribute could be a list of keys or a class type
          #
          if isinstance(intf[m],Box) and isinstance(mod_attr[m].link,list):
            for k in intf[m].keys():              # Iterate over node link-level module-specific attributes
              if not k in mod_attr[m].interface:  # If the name of an attribute is not in the list of allowed
                                                  # ... interface attributes report error
                common.error(
                  f"Node {n} has invalid attribute {k} for module {m} on link {l}",
                  common.IncorrectValue,
                  'modules')
          else:
            if type(intf[m]).__name__ != str(mod_attr[m].interface):
              common.error(
                f"Node {n} has invalid value type for attribute {k} for module {m} on link {l}, expected {mod_attr[m].interface}",
                common.IncorrectValue,
                'modules')

"""
Prepare module attribute dictionary

* If the module attributes are a list, then global/node/link/interface attributes are the same

Otherwise:

* Add propagatable link attributes to interface attributes
* Copy global attributes to node attributes if the node attributes are not specified
"""

def parse_module_attributes(a: typing.Union[typing.Dict, Box]) -> Box:
  if isinstance(a,dict):
    attr = Box(a,default_box=True,box_dots=True)
    if not isinstance(attr.interface,str):
      attr.interface = list(set(attr.link) - set(attr.link_no_propagate) | set(attr.interface))
    attr.node = attr.get("node",attr["global"])
  else:
    attr = Box({
      "global": a,
      "node": a,
      "link": a,
      "interface": a
    },default_box=True,box_dots=True)
  return attr

"""
check_module_dependencies:

For every module used by the topology, check is the module has "requires" attribute, and if
it does, check that everything in that list is also in the topology modules list.
"""
def check_module_dependencies(topology:  Box) -> None:
  for m in topology.get("module",[]):               # Use this format in case we don't use any modules
    mod_def = topology.defaults.get(m,{})           # Get module defaults
    if mod_def:                                     # Are they meaningful?
      if "requires" in mod_def:                     # Do they include list of required modules?
        for rqm in topology.defaults[m].requires:   # Loop over prerequisites
          if not rqm in topology.module:            # Now we can be explicit - we know topology.modules exists
            common.error("Module %s requires module %s which is not enabled in your topology" % (m,rqm))

"""
reorder_node_modules:

For every node with at least two modules: sort the list of modules based on their _requires_ and
_config_after_/_transform_after_ dependencies preserving the original order as much as possible.
"""

def reorder_node_modules(topology: Box, secondary_sort: str = "config_after") -> None:
  if 'module' in topology:
    topology.module = sort_module_list(topology.module,topology.defaults, secondary_sort)
    topology.defaults.module = topology.module

  for name,n in topology.nodes.items():
    if 'module' in n:
      n.module = sort_module_list(n.module,topology.defaults, secondary_sort)

def sort_module_list(mods: list, mod_params: Box, secondary_sort: str = "config_after") -> list:
  if (len(mods) < 2):
    return mods

  output: typing.List[str] = []
  while len(mods):
    skipped: typing.List[str] = []
    for m in mods:
      if m in mod_params:
        requires = mod_params[m].get('requires',[]) + mod_params[m].get(secondary_sort,[])
        if [ r for r in requires if r in mods ]:
          skipped = skipped + [ m ]
        else:
          output = output + [ m ]

    mods = skipped

  return output

"""
Copy node data into interface data:

For every module configured on a node, merge attributes listed in node_copy list
from node data to interface data. 

Example: copy node-level OSPF area into interfaces that don't have explicit area configuration.
"""

def copy_node_data_into_interfaces(topology: Box) -> None:
  for n in topology.nodes.values():
    for m in n.get('module',[]):                                 # Iterate over node modules
      if topology.defaults[m].attributes.node_copy:              # .. any copyable attributes for this module?
        copy_attr = Box({ k: v 
          for k,v in n.get(m,{}).items() 
            if k in topology.defaults[m].attributes.node_copy }) # Build a Box of node attributes that could be copied to interfaces
        if copy_attr:                                            # .. anything to copy?
          for intf in n.get('interfaces',[]):                    # .. if so, it would be nice to merge it with interface data
            if isinstance(intf.get(m,{}),dict):                  # .. but only if the interface data allows it
              intf[m] = copy_attr + intf[m]

"""
Callback transformation routines

* node_transform: for all nodes, call specified method for every module used by the node
* link_transform: for all links, call specified method for every module used by any node on the link

Note: mod_load is a global cache of loaded modules
"""

mod_load: typing.Dict = {}

def module_transform(method: str, topology: Box) -> None:
  global mod_load

  for m in topology.get('module',[]):
    if not mod_load.get(m):
      mod_load[m] = _Module.load(m,topology.get(m))
    mod_load[m].call("module_"+method,topology)

def node_transform(method: str , topology: Box) -> None:
  global mod_load

  for name,n in topology.nodes.items():
    for m in n.get('module',[]):
      if not mod_load.get(m):  # pragma: no cover (module should have been loaded already)
        mod_load[m] = _Module.load(m,topology.get(m))
      mod_load[m].call("node_"+method,n,topology)

def link_transform(method: str, topology: Box) -> None:
  global mod_load

  for l in topology.get("links",[]):
    mod_list: typing.Dict = {}
    for node_data in l.get(IFATTR,[]):
      mod_list.update({ m: None for m in topology.nodes[node_data.node].get("module",[]) })
    for m in mod_list.keys():
      if not mod_load.get(m):  # pragma: no cover (module should have been loaded already)
        mod_load[m] = _Module.load(m,topology.get(m))
      mod_load[m].call("link_"+method,l,topology)



#
# Create YAML or JSON output
#
import typing

import yaml
import os
from box import Box,BoxList

from .. import common
from .. import data
from ..augment import topology

from . import _TopologyOutput

class YAML(_TopologyOutput):

  def write(self, topo: Box) -> None:
    outfile = self.settings.filename or '-'
    modname = type(self).__name__

    if hasattr(self,'filenames'):
      outfile = self.filenames[0]
      if len(self.filenames) > 1:
        common.error('Extra output filename(s) ignored: %s' % str(self.filenames[1:]),common.IncorrectValue,modname)

    cleantopo: typing.Any = topology.cleanup_topology(topo)
    output = common.open_output_file(outfile)

    for fmt in self.format:
      if fmt == 'nodefault':
        cleantopo.pop('defaults')
      elif fmt == 'noaddr':
        cleantopo.pop('addressing')
      elif data.get_from_box(cleantopo,fmt):
        result = data.get_from_box(cleantopo,fmt)
        if not isinstance(result,Box) and not isinstance(result,BoxList):
          common.fatal(f'Selecting {fmt} did not result in a usable dictionary, aborting')
          return
        cleantopo = result
        break
      else:
        common.error('Invalid format modifier %s' % fmt,common.IncorrectValue,modname)

    common.exit_on_error()
    if modname == 'YAML':
      output.write(cleantopo.to_yaml())
    else:
      output.write(cleantopo.to_json(indent=2,sort_keys=True))
    if outfile != '-':
      common.close_output_file(output)
      print("Created transformed topology dump in YAML format in %s" % outfile)
    else:
      output.write("\n")



#
# Build full-blown topology data structures (nodes, links, global parameter) from high-level topology
#

import sys

from box import Box

from .. import common
from .. import addressing
from .. import augment
from ..providers import _Provider
from .. import modules

def transform_setup(topology: Box) -> None:
  augment.topology.check_required_elements(topology)
  topology.nodes = augment.nodes.create_node_dict(topology.nodes)
  if 'links' in topology:
    topology.links = augment.links.adjust_link_list(topology.links,topology.nodes)
    augment.links.set_linkindex(topology)
  augment.devices.augment_device_settings(topology)
  augment.groups.init_groups(topology)
  common.exit_on_error()

  augment.plugin.init(topology)
  augment.plugin.execute('init',topology)
  augment.topology.extend_attribute_list(topology.defaults)
  augment.topology.extend_module_attribute_list(topology)
  augment.topology.adjust_global_parameters(topology)
  topology.Provider = _Provider.load(topology.provider,topology.defaults.providers[topology.provider])
  common.exit_on_error()

  augment.nodes.augment_node_provider_data(topology)
  augment.nodes.augment_node_system_data(topology)
  common.exit_on_error()

def transform_data(topology: Box) -> None:
  addressing.setup(topology,topology.defaults)
  augment.plugin.execute('pre_transform',topology)
  modules.pre_transform(topology)

  augment.groups.adjust_groups(topology)
  augment.topology.check_global_elements(topology)

  augment.plugin.execute('pre_node_transform',topology)
  modules.pre_node_transform(topology)
  augment.nodes.transform(topology,topology.defaults,topology.pools)
  common.exit_on_error()
  augment.plugin.execute('post_node_transform',topology)

  if 'links' in topology:
    augment.plugin.execute('pre_link_transform',topology)
    modules.pre_link_transform(topology)
    augment.links.transform(topology.links,topology.defaults,topology.nodes,topology.pools)
    common.exit_on_error()
    augment.plugin.execute('post_link_transform',topology)

  modules.post_transform(topology)
  augment.plugin.execute('post_transform',topology)
  augment.groups.node_config_templates(topology)
  topology.Provider.call("post_transform",topology)
  common.exit_on_error()

  topology.pop('Plugin',None)
  del topology.pools
  del topology.Provider

def transform(topology: Box) -> None:
  transform_setup(topology)
  transform_data(topology)
